"""
Miscellaneous mixin classes for object in OpenTNSim.

The following classes are provided:
- Neighbours
- ExtraMetadata

"""
# package(s) for documentation, debugging, saving and loading
import logging
from typing import Union

# math packages
import numpy as np

# get logger
logger = logging.getLogger(__name__)


class Neighbours:
    """Mixin class: Can be added to a locatable object (list)

    Parameters
    ----------
    travel_to: list
            list of locatables to which can be travelled
    args, kwargs:
        passed to SimpyObject. Must at least contain parameter env: simpy.Environment.

    Attributes
    -----------
    neighbours: list
        list of locatables to which can be travelled.
    """

    def ___init(self, travel_to: list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """Initialization"""
        self.neighbours = travel_to


class ExtraMetadata:
    """Mixin class: store all leftover keyword arguments as metadata property (use as last mixin)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        # store all other properties as metadata
        self.metadata = kwargs



from .lock import (
    PassesLockComplex,
    IsLockComplex,
    IsLockMaster,
    HasLockPlanning,
    IsLockWaitingArea,
)
from .lock_chamber import IsLockChamber, IsLockChamberOperator
from .calculations import levelling_time_equation
from .logutils import calculate_cycle_looptimes, calculate_detailed_cycle_time, get_vessels_during_leveling

__all__ = [
    "PassesLockComplex",
    "IsLockComplex",
    "IsLockMaster",
    "HasLockPlanning",
    "IsLockWaitingArea",
    "IsLockChamber",
    "IsLockChamberOperator",
    "levelling_time_equation",
    "calculate_cycle_looptimes",
    "calculate_detailed_cycle_time",
    "get_vessels_during_leveling",
]



"""Contains the mixin for lock chambers. Also contains the parent class LockChamberOperatior."""
import simpy
import numpy as np
import pandas as pd
import datetime
import math
import functools

from opentnsim.core import HasResource, Identifiable, Log, HasLength, ExtraMetadata
from opentnsim.utils import time_to_numpy
from opentnsim.lock.calculations import calculate_z, levelling_time_equation
from opentnsim.lock.utils import _get_lock_operation_to_and_from_node
from opentnsim.vessel_traffic_service.hydrodanamic_data_manager import HydrodynamicDataManager
from opentnsim.output import HasOutput
from opentnsim.graph.mixins import HasMultiDiGraph, get_length_of_edge
from opentnsim.constants import knots


class IsLockChamberOperator:
    """The lock chamber operator operates one chamber of the lock.

    The operator communicates with the lock master through self.lock_master
    to coordinate vessel operations and lock state changes.
    """

    def __init__(self, lock_master, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock_master = lock_master

    def _get_appropriate_waiting_area(self, direction):
        """Get appropriate waiting area through lock master.

        Parameters
        ----------
        direction : int
            the direction of the vessel: 0 (A -> B) or 1 (B -> A)
        """
        if self.lock_master:
            return self.lock_master._get_appropriate_waiting_area(direction)
        else:
            raise AttributeError("No lock master set for accessing waiting areas")

    @property
    def closing_doors_in_between_arrivals(self):
        """Get policy on closing doors in between arrivals through lock master."""
        if self.lock_master:
            return self.lock_master.closing_doors_in_between_arrivals
        else:
            raise AttributeError("No lock master set for accessing door closing policy")

    @property
    def closing_doors_in_between_operations(self):
        """Get policy on closing doors in between operations through lock master."""
        if self.lock_master:
            return self.lock_master.closing_doors_in_between_operations
        else:
            raise AttributeError("No lock master set for accessing door closing policy")

    def close_doors_before_vessel_is_laying_still(self, operation_index):
        """Get policy on closing doors before vessel is laying still through lock master.

        Parameters
        ----------
        operation_index : int
            index of the lock operation
        """
        if self.lock_master:
            return self.lock_master.close_doors_before_vessel_is_laying_still(operation_index)
        else:
            raise AttributeError("No lock master set for accessing door closing before vessel laying still policy")

    @property
    def _distance_to_lock(self):
        """Get distance to lock through lock master."""
        if self.lock_master:
            return self.lock_master._distance_to_lock
        else:
            raise AttributeError("No lock master set for accessing distance to lock")

    @property
    def vessel_planning(self):
        """Get vessel planning through lock master."""
        if self.lock_master:
            return self.lock_master.vessel_planning
        else:
            raise AttributeError("No lock master set for accessing vessel planning")

    @property
    def operation_planning(self):
        """Get operation planning through lock master."""
        if self.lock_master:
            return self.lock_master.operation_planning
        else:
            raise AttributeError("No lock master set for accessing operation planning")

    @property
    def waiting_area_B(self):
        """Get waiting area B through lock master."""
        if self.lock_master:
            return self.lock_master.waiting_area_B
        else:
            raise AttributeError("No lock master set for accessing waiting area B")

    @property
    def waiting_area_A(self):
        """Get waiting area A through lock master."""
        if self.lock_master:
            return self.lock_master.waiting_area_A
        else:
            raise AttributeError("No lock master set for accessing waiting area A")

    def register_vessel(self, vessel):
        """Register vessel through lock master.

        Parameters
        ----------
        vessel : type
            a type including the following mixins: PassesLockComplex,Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        """
        if self.lock_master:
            yield from self.lock_master.register_vessel(vessel)

    def calculate_sailing_time_to_waiting_area(self, vessel, direction, current_node=None, prognosis=False, overwrite=True):
        """Calculate sailing time to waiting area through lock master.

        Parameters
        ----------
        vessel : type
            a type including the following mixins: PassesLockComplex,Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        direction : int
            the direction of the vessel: 0 (A -> B) or 1 (B -> A)
        current_node : str
            node name (that has to be in the graph) on which the vessel is currently sailing, to navigate an edge should form an edge with the origin)
        prognosis : bool
            if the sailing time is calculated for prognosis purposes (True) or for actual sailing (False)
        overwrite : bool
            if existing sailing time in the vessel planning should be overwritten (True) or not (False)
        """

        if self.lock_master:
            return self.lock_master.calculate_sailing_time_to_waiting_area(vessel, direction, current_node, prognosis, overwrite)

    def calculate_vessel_departure_start_delay(self, vessel, operation_index, direction, prognosis=False):
        """Calculate vessel departure start delay through lock master.

        Parameters
        ----------
        vessel : type
            a type including the following mixins: PassesLockComplex,Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        operation_index : int
            index of the lock operation
        direction : int
            the direction of the vessel: 0 (A -> B) or 1 (B -> A)
        prognosis : bool
            if the delay is calculated for prognosis purposes (True) or for actual sailing (False)
        overwrite : bool
            if existing delay in the vessel planning should be overwritten (True) or not (False)
        """
        if self.lock_master:
            return self.lock_master.calculate_vessel_departure_start_delay(vessel, operation_index, direction, prognosis)

    def initiate_levelling(self, origin, destination, vessel=None, k=0):
        """
        Initiates levelling process as function that can be added to a vessel TODO: preferably you don't want to add this process to the vessel but let the lock master / operator handle this

        Parameters
        ----------
        origin : str
            node name (that has to be in the graph) on which the vessel is currently sailing, to navigate an edge should form an edge with the origin)
        destination : str
            node name (that has to be in the graph) on which the vessel is currently sailing to, to navigate an edge (should form an edge with the origin)
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        k : int
            identifier of the edge between two nodes in a multidigraph network


        """
        # TODO: Moeten de origin en destination hier naast elkaar liggen? Zoja, toevoegen in documentatie.

        # determine if there is a lock on the edge
        if "Lock" not in vessel.multidigraph.edges[origin, destination, k].keys():
            return

        # get the lock complex object
        lock = vessel.multidigraph.edges[origin, destination, k]["Lock"][0]

        # unpack the lock complex master's vessel and lock operation plannings
        vessel_planning = lock.vessel_planning
        operation_planning = lock.operation_planning

        # determine the index of the vessel and the lock operation to which it is assigned to and the index of this operation
        vessel_planning_index = vessel_planning[vessel_planning.id == vessel.id].iloc[-1].name
        operation_index = vessel_planning.loc[vessel_planning_index, "operation_index"]
        this_operation = operation_planning.loc[operation_index]

        # determine the direction to the lock chamber is currently levelled to, and to which node the lock chamber will level
        current_node = lock.node_open
        if current_node == lock.start_node:
            direction = 0
            next_node = lock.end_node
        else:
            direction = 1
            next_node = lock.start_node

        # determine the vessels that are assigned to the lock operation to which the vessel is assigned
        vessels = this_operation.vessels

        # initiate levelling if vessel is the last assigned vessel in the lock
        if vessel == vessels[-1]:
            # liberate the vessels that were requested to wait for the last vessel
            for other_vessel in vessels[:-1]:
                terminate_waiting_time_for_other_vessel = False
                while not terminate_waiting_time_for_other_vessel:
                    try:
                        yield lock.wait_for_other_vessels.put(other_vessel)
                        terminate_waiting_time_for_other_vessel = True
                    except simpy.Interrupt as e:
                        terminate_waiting_time_for_other_vessel = False

            # Wait for other vessels to lay still
            delay = (operation_planning.loc[operation_index].time_door_closing_start.round("s").to_pydatetime().timestamp() - lock.env.now)
            if delay > 0:
                yield lock.env.timeout(delay)

            # Convert lock chamber
            close_doors = True
            if (lock.close_doors_before_vessel_is_laying_still
                and this_operation.time_door_closing_start < vessel_planning.loc[vessel_planning_index, "time_lock_entry_stop"]):
                close_doors = False

            lock.operation_planning.loc[operation_index, 'status'] = 'unavailable'
            yield from lock.convert_chamber(next_node, direction, operation_index=operation_index, vessel=vessel, close_doors=close_doors)

            # Liberate waiting vessels in lock chamber
            for other_vessel in vessels[:-1]:
                terminate_levelling_for_other_vessel = False
                while not terminate_levelling_for_other_vessel:
                    try:
                        yield lock.wait_for_levelling.put(other_vessel)
                        terminate_levelling_for_other_vessel = True
                    except simpy.Interrupt as e:
                        terminate_levelling_for_other_vessel = False

        # If vessel is not the last assigned vessel
        else:
            # Wait for last assigned vessel of lock operation
            waiting_for_other_vessels = True
            last_location = vessel.logbook[-1]["Geometry"]
            vessel.log_entry_v0("Waiting for other vessels in lock start", self.env.now, self.output.copy(), last_location)
            while waiting_for_other_vessels:
                try:
                    yield lock.wait_for_other_vessels.get(filter=(lambda request: request.id == vessel.id))
                    waiting_for_other_vessels = False
                except simpy.Interrupt as e:
                    waiting_for_other_vessels = True
            vessel.log_entry_v0("Waiting for other vessels in lock stop", self.env.now, self.output.copy(),last_location)

            # Follow the converting lock chamber
            vessel.log_entry_v0(
                "Levelling start",
                vessel.env.now,
                vessel.output.copy(),
                vessel.position_in_lock,
            )
            waiting_for_levelling = True
            while waiting_for_levelling:
                try:
                    yield lock.wait_for_levelling.get(filter=(lambda request: request.id == vessel.id))
                    waiting_for_levelling = False
                except simpy.Interrupt as e:
                    waiting_for_levelling = True
            vessel.log_entry_v0(
                "Levelling stop",
                vessel.env.now,
                vessel.output.copy(),
                vessel.position_in_lock,
            )

        # determine and yield sailing out delay
        sailing_out_delay = lock.calculate_vessel_departure_start_delay(vessel, operation_index, direction).total_seconds()
        delay_start = vessel.env.now
        while sailing_out_delay:
            try:
                yield vessel.env.timeout(sailing_out_delay)
                sailing_out_delay = 0
            except simpy.Interrupt as e:
                sailing_out_delay -= vessel.env.now - delay_start

    def prepare_next_lock_operation(self, lock, operation_index, direction, vessel):
        """Lock operator checks and (if required) initiates an empty lock operation or closes the doors if there is sufficient time with respect to the next operation's start time

        Parameters
        ----------
        lock : object
            the lock chamber object generated with IsLockChamber
        operation_index : int
            index of the lock operation
        direction : int
            the direction of the lock operation: 0 (A -> B) or 1 (B -> A)
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        """
        # get variables of the last lock operation: do nothing if it is not the last vessel that is sailing out of the lock
        operation_planning = self.lock_master.operation_planning
        last_operation = operation_planning.loc[operation_index]
        vessels_in_last_operation = last_operation.vessels
        is_last_vessel_sailing_out = vessels_in_last_operation[-1] == vessel
        if not is_last_vessel_sailing_out:
            return

        # get the current time, and the information of the next operation
        current_time = pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now))
        _, to_node = _get_lock_operation_to_and_from_node(self, 1 - direction)
        next_operations = operation_planning[operation_planning.index > operation_index]

        # determine if the doors can be closed after the considered vessel has sailed out of the lock
        doors_can_be_closed = lock.determine_if_door_can_be_closed(vessel, direction, operation_index)

        # determine if the next operation is empty
        next_lockage_is_empty = False
        if not next_operations.empty:
            next_operation = next_operations.iloc[0]
            if not len(next_operation.vessels):
                next_lockage_is_empty = True

        # an action should be done if the doors can be closed in between operations, or if the next lock operation is empty
        if doors_can_be_closed and lock.closing_doors_in_between_operations:
            door_closing_start_time = last_operation.time_potential_lock_door_closure_start
            delay = np.max([self.sailing_time_before_closing_lock_doors, (door_closing_start_time - current_time).total_seconds()])

            # close the doors with the correct delay
            vessel.env.process(lock.close_door(delay=delay))

        elif next_lockage_is_empty:
            door_closing_start_time = next_operation.time_door_closing_start
            closing_delay = np.max([self.sailing_time_before_closing_lock_doors, (door_closing_start_time - current_time).total_seconds()])

            # if there is an empty lock operation and no policy that doors are closed in between operations is active -> close doors and convert chamber afterwards
            if not lock.closing_doors_in_between_operations:
                convert_chamber_delay = closing_delay
                closing_doors = True
            # if there is an empty lock operation but the policy that doors are closed in between operations is active -> close doors and convert chamber later, or convert chamber immediately if there is insufficient time
            else:
                next_operation = next_operations.iloc[1]
                door_opening_start_time = next_operation.time_potential_lock_door_opening_stop
                lock_operation_duration = self.determine_time_to_open_door(operation_index = vessel_operation_index + 1,
                                                                           direction =1 - direction,
                                                                           doors_required_to_be_open = door_opening_start_time)
                opening_delay = (np.max([0, (door_opening_start_time - current_time).total_seconds()]) - lock_operation_duration.total_seconds())
                if opening_delay > (closing_delay + self.lock_master.doors_closing_time):
                    convert_chamber_delay = opening_delay
                    closing_doors = False
                    vessel.env.process(lock.close_door(delay=closing_delay))
                else:
                    convert_chamber_delay = closing_delay
                    closing_doors = True

            # convert the lock chamber with the correct delay and if the doors should first be closed
            vessel.env.process(lock.convert_chamber(operation_index = operation_index + 1,
                                                    new_level = to_node,
                                                    vessel = None,
                                                    close_doors = closing_doors,
                                                    delay = convert_chamber_delay,
                                                    direction = 1 - direction))

    def allow_vessel_to_sail_out_of_lock(self, origin, destination, vessel=None, k=0):
        """Allows the vessel to sail out of the lock chamber

        Parameters
        ----------
        origin : str
            node name (that has to be in the graph) on which the vessel is currently sailing, to navigate an edge should form an edge with the origin)
        destination : str
            node name (that has to be in the graph) on which the vessel is currently sailing to, to navigate an edge (should form an edge with the origin)
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        k : int
            identifier of the edge between two nodes in a multidigraph network

        Yields
        ------
        Vessel to sail to the end of the edge at which the lock chamber is located, and initiates new processes: i.e. closing doors or empty lock operation
        """
        # checks if lock is present on the edge
        if "Lock" not in vessel.multidigraph.edges[origin, destination, k].keys():
            return

        # unpacks the lock and vessel and operation planning
        lock = vessel.multidigraph.edges[origin, destination, k]["Lock"][0]
        vessel_planning = lock.vessel_planning

        # determines information of the lock operation
        vessel_planning_index = vessel_planning[vessel_planning.id == vessel.id].iloc[-1].name
        operation_index = vessel_planning.loc[vessel_planning_index, "operation_index"]
        direction = vessel_planning.loc[vessel_planning_index, "direction"]

        # determines the distance from the vessel to the lock doors that have to be passed
        distance_in_lock_from_position = lock.lock_length - vessel.distance_position_from_first_lock_doors

        # determines the geometry objects of the lock based on the direction of the vessel TODO: function?
        if not direction:
            second_lock_doors_position = lock.location_lock_doors_B
            remaining_distance = lock.distance_from_end_node_to_lock_doors_B
            exit_geom = vessel.env.graph.nodes[lock.end_node]["geometry"]
        else:
            second_lock_doors_position = lock.location_lock_doors_A
            remaining_distance = lock.distance_from_start_node_to_lock_doors_A
            exit_geom = vessel.env.graph.nodes[lock.start_node]["geometry"]

        # releasing the length of the lock TODO: only the yield statement should be kept: now it is prevented that vessels need to wait to put back their length, but in principle this should not occur although it sometimes occurs due to bugs
        release_lock_access = False
        while not release_lock_access:
            try:
                yield lock.length.put(vessel.L)
                release_lock_access = True
            except simpy.Interrupt as e:
                release_lock_access = True

        # determine the waiting time to sail out of the lock TODO: have another algorithm to determine this time: now the vessel planning is used, but this should be prevented -> the lock planning might be off by a few seconds/minutes due to uncertainties/errors in predictions and unforeseen circumstances
        waiting_to_sail_out_time = (vessel_planning.loc[vessel_planning_index, "time_lock_departure_start"] -
                                    pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now))).total_seconds()

        # let the vessel wait to sail out of the lock (vessels may have to give other vessels priority to sail out to later sail out of the lock in a safe manner with sufficient distance to the vessel ahead -> i.e., if they sailed into the lock ahead of the considered vessel blocking the sailing out path)
        waiting_to_sail_out_time_start = vessel.env.now
        while waiting_to_sail_out_time > 0:
            try:
                yield vessel.env.timeout(waiting_to_sail_out_time)
                waiting_to_sail_out_time = 0
            except simpy.Interrupt as e:
                waiting_to_sail_out_time -= vessel.env.now - waiting_to_sail_out_time_start

        # log that the vessel can start sailing out of the lock (up to the lock doors)
        vessel.log_entry_v0("Sailing to second lock doors start", vessel.env.now, vessel.output.copy(), vessel.position_in_lock,)

        # determine the process of sailing to the lock doors that have to be passed (distance to these doors divided by the sailing out speed of the vessel)
        vessel_speed = lock.vessel_sailing_speed_out_lock(vessel)
        sailing_out_time = distance_in_lock_from_position / vessel_speed
        sailing_out_start = vessel.env.now
        while sailing_out_time:
            try:
                yield vessel.env.timeout(sailing_out_time)
                sailing_out_time = 0
            except simpy.Interrupt as e:
                sailing_out_time -= vessel.env.now - sailing_out_start

        # log that the vessel can stops sailing out of the lock (up to the lock doors)
        vessel.log_entry_v0("Sailing to second lock doors stop", vessel.env.now, vessel.output.copy(), second_lock_doors_position,)

        # remove functions specific to passing the lock chamber
        remove_functions = [lock.allow_vessel_to_sail_into_lock, lock.initiate_levelling, lock.allow_vessel_to_sail_out_of_lock]
        remove_on_pass_edge_functions = []
        for index, function in enumerate(vessel.on_pass_edge_functions):
            if isinstance(function, functools.partial):
                if function.func in remove_functions:
                    remove_on_pass_edge_functions.append(function)
            elif function in remove_functions:
                remove_on_pass_edge_functions.append(function)
        for function in remove_on_pass_edge_functions:
            vessel.on_pass_edge_functions.remove(function)

        # determine if the lock has to be levelled
        self.prepare_next_lock_operation(lock, operation_index, direction, vessel)

        # log that sailing out of the lock complex is starting
        vessel.log_entry_v0("Sailing to lock complex exit start", vessel.env.now, vessel.output.copy(), second_lock_doors_position)

        # let the vessel sail to the end of the lock complex
        vessel_speed = lock.vessel_sailing_out_speed(vessel, direction)
        sailing_out_time = remaining_distance / vessel_speed
        sailing_out_start = vessel.env.now
        while sailing_out_time:
            try:
                yield vessel.env.timeout(sailing_out_time)
                sailing_out_time = 0
            except simpy.Interrupt as e:
                sailing_out_time -= vessel.env.now - sailing_out_start
                remaining_sailing_distance = vessel_speed * sailing_out_time
                sailing_out_time = remaining_sailing_distance / vessel.current_speed

        # log that sailing out of the lock complex is stopping and set that no distance has to be sailed along the edge (vessel is at end of lock complex)
        vessel.log_entry_v0("Sailing to lock complex exit stop", vessel.env.now, vessel.output.copy(), exit_geom,)
        vessel.distance_left_on_edge = 0

    def allow_vessel_to_sail_into_lock(self, origin, destination, vessel=None, k=0):
        """Allows the vessel to sail into the lock chamber

        Parameters
        ----------
        origin : str
            node name (that has to be in the graph) on which the vessel is currently sailing, to navigate an edge should form an edge with the origin)
        destination : str
            node name (that has to be in the graph) on which the vessel is currently sailing to, to navigate an edge (should form an edge with the origin)
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        k : int
            identifier of the edge between two nodes in a multidigraph network
        """

        # checks if lock is present on the edge
        if "Lock" not in vessel.multidigraph.edges[origin, destination, k].keys():
            return

        # unpacks the lock and vessel and operation planning
        lock = vessel.multidigraph.edges[origin, destination, k]["Lock"][0]
        vessel_planning = lock.vessel_planning

        # determines information of the lock operation
        vessel_planning_index = vessel_planning[vessel_planning.id == vessel.id].iloc[-1].name
        operation_index = vessel_planning.loc[vessel_planning_index, "operation_index"]
        direction = vessel_planning.loc[vessel_planning_index, "direction"]
        current_time = vessel.env.now

        # determines the geometry objects of the lock based on the direction of the vessel TODO: function?
        waiting_area = lock._get_appropriate_waiting_area(direction)
        distance_to_lock = lock._distance_to_lock(direction)
        if not direction:
            first_lock_door_position = lock.location_lock_doors_A
        else:
            first_lock_door_position = lock.location_lock_doors_B

        # correct the distance to the lock doors if the vessel is in the waiting area, located at the same edge of the lock
        lock_start_node, lock_end_node = _get_lock_operation_to_and_from_node(self, direction)
        if (lock_start_node, lock_end_node) == waiting_area.edge:
            distance_to_lock -= waiting_area.distance_from_edge_start

        # log the start of sailing to the lock doors
        last_position_vessel = vessel.logbook[-1]["Geometry"]
        vessel.log_entry_v0("Sailing to first lock doors start", vessel.env.now, vessel.output.copy(), last_position_vessel,)

        # let vessel sail to the lock doors
        vessel_speed = lock.vessel_sailing_in_speed(vessel, direction)
        remaining_sailing_time = distance_to_lock / vessel_speed
        while remaining_sailing_time > 0:
            try:
                yield vessel.env.timeout(remaining_sailing_time)
                remaining_sailing_time = 0
            except simpy.Interrupt as e:
                remaining_sailing_time -= vessel.env.now - current_time
                remaining_sailing_distance = vessel_speed * remaining_sailing_time
                remaining_sailing_time = remaining_sailing_distance / vessel.current_speed

        # vessel entering now the lock -> delete the overruled speeds imposed on the vessel
        vessel.overruled_speed = vessel.overruled_speed.iloc[0:0]

        # claim the lock length (this should not lead to waiting time)
        yield lock.length.get(vessel.L)

        # log the stop of sailing to the lock doors
        vessel.log_entry_v0("Sailing to first lock doors stop", vessel.env.now, vessel.output.copy(), first_lock_door_position,)

        # Checks if door should be closed intermediately
        vessel_planning_index = vessel_planning[vessel_planning.id == vessel.id].iloc[-1].name

        # calculate delay to close doors
        current_time = pd.Timestamp(datetime.datetime.fromtimestamp(vessel.env.now))
        delay_to_close_doors = vessel_planning.loc[vessel_planning_index, "time_potential_lock_door_closure_start"] - current_time

        # close doors if doors can be closed in between vessel arrivals or if vessel is last vessel to enter the lock
        doors_can_be_closed_between_vessel_arrivals = lock.determine_if_door_can_be_closed(vessel, direction, operation_index, between_arrivals=True)
        if lock.close_doors_before_vessel_is_laying_still and doors_can_be_closed_between_vessel_arrivals:
            vessel.env.process(lock.close_door(delay=delay_to_close_doors.total_seconds()))

        # log the start of sailing to the position within the lock chamber
        vessel.log_entry_v0("Sailing to position in lock start", vessel.env.now, vessel.output.copy(), first_lock_door_position, )

        # determine position in the lock chamber and distance to sail to this location
        vessel.distance_position_from_first_lock_doors = lock.length.level + 0.5 * vessel.L
        if not direction:
            distance_to_position_in_lock = lock.distance_from_start_node_to_lock_doors_A + vessel.distance_position_from_first_lock_doors
        else:
            distance_to_position_in_lock = lock.distance_from_end_node_to_lock_doors_B + vessel.distance_position_from_first_lock_doors
        vessel.position_in_lock = vessel.env.vessel_traffic_service.provide_location_over_edges(lock_start_node, lock_end_node, distance_to_position_in_lock)

        # let vessel sail to the assigned location in the lock chamber
        vessel_speed = lock.vessel_sailing_speed_in_lock(vessel)
        remaining_sailing_time = vessel.distance_position_from_first_lock_doors / vessel_speed
        while remaining_sailing_time > 0:
            try:
                yield vessel.env.timeout(remaining_sailing_time)
                remaining_sailing_time = 0
            except simpy.Interrupt as e:
                remaining_sailing_time -= vessel.env.now - start_sailing

        # log the stop of the sailing event to the assigned locaiton in the lock chamber
        vessel.log_entry_v0("Sailing to position in lock stop", vessel.env.now, vessel.output.copy(), vessel.position_in_lock,)

        # close doors if doors can be closed between vessel arrivals and doors have not already been closed before
        doors_can_be_closed_between_vessel_arrivals = lock.determine_if_door_can_be_closed(vessel, direction, operation_index, between_arrivals=True)
        if not lock.close_doors_before_vessel_is_laying_still and doors_can_be_closed_between_vessel_arrivals:
            vessel.env.process(lock.close_door())

    def convert_chamber(self, new_level, direction, operation_index=None, vessel=None, close_doors=True, delay=0.0):
        """
        Converts the lock chamber and logs this event. TODO: attribute for lock operator

        Parameters
        ----------
        new_level : str
            node that represents the side at which the lock is currently levelled
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        close_doors : bool
            if the doors have to be closed: yes (True) or no (False)
        delay : float
            a delay before lock conversion [s]

        direction : int
            the direction of the vessel: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)

        Yields
        ------
        The conversion of the lock chamber
        """
        if operation_index is not None:
            self.lock_master.operation_planning.loc[operation_index, "status"] = "unavailable"

        # if there is a delay -> yield time out
        start_delay = self.env.now
        while delay > 0:
            try:
                yield self.env.timeout(delay)
                delay = 0
            except simpy.Interrupt as e:
                delay -= self.env.now - start_delay

        # close the doors or make sure that lock is not performing another process
        if close_doors:
            yield from self.close_door(delay=delay)
        else:
            hold_door_A = self.door_A.request()
            hold_levelling = self.levelling.request()
            hold_door_B = self.door_B.request()
            yield hold_door_A
            yield hold_levelling
            yield hold_door_B
            self.door_A.release(hold_door_A)
            self.levelling.release(hold_levelling)
            self.door_B.release(hold_door_B)

        # level lock and open the doors afterwards
        yield from self.level_lock(new_level, direction, operation_index=operation_index, vessel=vessel)
        yield from self.open_door()

    def close_door(self, delay=0.0):
        """
        Lock operator closes the lock doors TODO: attribute for lock operator

        Parameters
        ----------
        delay : float
            a delay before door opening [s]

        Yields
        ------
        The closing of the door
        """

        # if there is a delay -> yield time out
        start_delay = self.env.now
        while delay > 0:
            try:
                yield self.env.timeout(delay)
                delay = 0
            except simpy.Interrupt as e:
                delay -= self.env.now - start_delay

        # make sure that all lock elements are requested, so only one process is occurring
        hold_door_A = self.door_A.request()
        hold_levelling = self.levelling.request()
        hold_door_B = self.door_B.request()
        yield hold_door_A
        yield hold_levelling
        yield hold_door_B

        # log the start of the event
        self.log_entry_v0("Lock doors closing start", self.env.now, self.output.copy(), self.node_open)

        # timeout event of the doors closing
        remaining_doors_closing_time = self.doors_closing_time
        start_time_closing = self.env.now
        while remaining_doors_closing_time:
            try:
                yield self.env.timeout(remaining_doors_closing_time)
                remaining_doors_closing_time = 0
            except simpy.Interrupt as e:
                remaining_doors_closing_time -= self.env.now - start_time_closing

        # set water level to the side at which the door has been closed
        time = np.datetime64(datetime.datetime.fromtimestamp(self.env.now))
        if self.node_open == self.start_node:
            node = self.start_node
        else:
            node = self.end_node
        time_index = HydrodynamicDataManager()._get_time_index_of_hydrodynamic_data(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time
        )
        new_water_level = HydrodynamicDataManager()._get_hydrodynamic_data_value(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time, node, "Water level"
        )
        self.water_level[time_index:] = new_water_level

        # log the end of the event
        self.log_entry_v0("Lock doors closing stop", self.env.now, self.output.copy(), self.node_open)
        if self.node_open == self.start_node:
            self.door_A_open = False
        else:
            self.door_B_open = False

        # release all lock elements that were requested, so the next process can start
        self.door_A.release(hold_door_A)
        self.levelling.release(hold_levelling)
        self.door_B.release(hold_door_B)

    def level_lock(self, new_level, direction, operation_index=None, vessel=None):
        """
        Lock operator levels the water level of the lock chamber to the harbour side of the direction of the lock operation TODO: attribute for lock operator

        new_level : str
            node of the edge of lock complex to which the lock chamber is levelling
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        direction : int
            the direction of the vessel: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        same_direction : bool


        Yields
        ------
        Levelling of the lock chamber
        """
        # make sure that all lock elements are requested, so only one process is occurring
        hold_door_A = self.door_A.request()
        hold_levelling = self.levelling.request()
        hold_door_B = self.door_B.request()
        yield hold_door_A
        yield hold_levelling
        yield hold_door_B

        # determine the levelling time
        levelling_time, _, _ = self.determine_levelling_time(t_start=self.env.now, direction=direction, operation_index=operation_index)

        # log the start of the event
        if vessel is not None:
            vessel.log_entry_v0(
                "Levelling start",
                vessel.env.now,
                vessel.output.copy(),
                vessel.position_in_lock,
            )
        self.log_entry_v0(
            "Lock chamber converting start",
            self.env.now,
            self.output.copy(),
            self.node_open,
        )

        # set new node to which the doors will be opened
        self.node_open = new_level

        # timeout
        remaining_levelling_time = levelling_time
        start_levelling = self.env.now
        while remaining_levelling_time:
            try:
                yield self.env.timeout(remaining_levelling_time)
                remaining_levelling_time = 0
            except simpy.Interrupt as e:
                remaining_levelling_time -= self.env.now - start_levelling

        # log the end of the event
        self.log_entry_v0(
            "Lock chamber converting stop",
            self.env.now,
            self.output.copy(),
            self.node_open,
        )
        if vessel is not None:
            vessel.log_entry_v0(
                "Levelling stop",
                vessel.env.now,
                vessel.output.copy(),
                vessel.position_in_lock,
            )

        # release all lock elements that were requested, so the next process can start
        self.door_A.release(hold_door_A)
        self.levelling.release(hold_levelling)
        self.door_B.release(hold_door_B)

    def open_door(self, to_level=None, vessel=None, delay=0.0):
        """
        Lock operator opens the lock doors TODO: attribute for lock operator

        Parameters
        ----------
        to_level : str
            node of the edge of lock complex to which the lock chamber opens
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        delay : float
            a delay before door opening

        Yields
        ------
        The opening of the door
        """

        # if there is a delay -> yield time out
        start_delay = self.env.now
        while delay > 0:
            try:
                yield self.env.timeout(delay)
                delay = 0
            except (
                simpy.Interrupt
            ) as e:  # if there is a delay -> yield time out with new delay (remaining delay added with a delay equal to the exception)
                delay -= self.env.now - start_delay
                if vessel is not None:
                    if e.cause is not None:
                        delay += float(e.cause)

        # delete attribute as form of communication of the vessel TODO: a bit complex, better do it in another way
        if vessel is not None:
            delattr(vessel, "door_open_request")

        hydromanager = HydrodynamicDataManager()

        # determine the water level in the lock chamber
        time = np.datetime64(datetime.datetime.fromtimestamp(self.env.now))
        time_index = hydromanager._get_time_index_of_hydrodynamic_data(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time
        )
        wlev_chamber = self.water_level[time_index]

        # determine to_level
        if to_level is None:
            to_level = self.node_open

        # determine the water level in the harbour
        wlev_harbour = hydromanager._get_hydrodynamic_data_value(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time, to_level, "Water level"
        )

        # determine the direction to which the vessels are sailing out
        if to_level == self.start_node:
            direction = 1
        else:
            direction = 0

        # if the water levels in the chamber and harbour are not aligned -> level lock again
        if not math.isnan(wlev_chamber) and not math.isnan(wlev_harbour) and np.abs(wlev_chamber - wlev_harbour) >= 0.1:
            yield from self.level_lock(to_level, direction=direction)
        else:
            self.node_open = to_level

        time = np.datetime64(datetime.datetime.fromtimestamp(self.env.now))
        time_index = hydromanager._get_time_index_of_hydrodynamic_data(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time
        )
        wlev_series_node_door_open = hydromanager._get_hydrodynamic_data_series(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time, self.node_open, "Water level"
        )
        self.water_level[time_index:] = wlev_series_node_door_open

        # make sure that all lock elements are requested, so only one process is occurring
        hold_door_A = self.door_A.request()
        hold_levelling = self.levelling.request()
        hold_door_B = self.door_B.request()
        yield hold_door_A
        yield hold_levelling
        yield hold_door_B

        # log the process start
        self.log_entry_v0("Lock doors opening start", self.env.now, self.output.copy(), self.node_open)

        # timeout
        remaining_doors_opening_time = self.doors_opening_time
        start_time_opening = self.env.now
        while remaining_doors_opening_time:
            try:
                yield self.env.timeout(remaining_doors_opening_time)
                remaining_doors_opening_time = 0
            except simpy.Interrupt as e:
                remaining_doors_opening_time -= self.env.now - start_time_opening

        # log the process stop
        self.log_entry_v0(
            "Lock doors opening stop",
            self.env.now,
            self.output.copy(),
            self.node_open,
        )

        # determine which side the door is open to
        if self.node_open == self.start_node:
            self.door_A_open = True
        else:
            self.door_B_open = True

        # release all lock elements that were requested, so the next process can start
        self.door_A.release(hold_door_A)
        self.levelling.release(hold_levelling)
        self.door_B.release(hold_door_B)

    def minimum_delay_to_close_doors(self):
        """
        Calculates the time delay (in seconds) between when the last vessel has entered the lock and when the lock doors can be closed

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput

        Returns
        -------
        minimum_delay_to_close_doors : pd.Timedelta
            the minimum time delay that the lock doors can be closed after a vessel has entered the lock
        """
        minimum_delay_to_close_doors = pd.Timedelta(seconds=self.sailing_time_before_closing_lock_doors)
        return minimum_delay_to_close_doors

    def minimum_advance_to_open_doors(self):
        """
        Determines the minimum time in advance that a lock door should be opened

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        direction : int
            the direction of the vessel: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)


        Returns
        -------
        minimum_advance_to_open_doors : pd.Timedelta
            the minimum time in advance that a lock door should be opened [s]

        """
        minimum_advance_to_open_doors = pd.Timedelta(seconds=self.sailing_time_before_opening_lock_doors)
        # minimum_advance_to_open_doors += pd.Timedelta(seconds=vessel.L/self.vessel_sailing_in_speed(vessel,direction))
        # TODO: take into account the vessels' bows and sterns to determine the time before and after which the door should be respectively opened and closed
        return minimum_advance_to_open_doors

    def determine_if_door_can_be_closed(self, vessel, direction, operation_index, between_arrivals=False):
        """
        Determines if the doors can be closed in between operations or vessel arrivals

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        operation_index : int
            index of the lock operation
        direction : int
            the direction of the lock operation: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        between_arrivals : bool
            if the function is run to determine if the doors can be closed in between vessel arrivals (True) or not (False)

        Returns
        -------
        doors_can_be_closed : bool
            doors can be closed (True) or not (False)
        """
        operation_planning = self.lock_master.operation_planning
        this_operation = operation_planning.loc[operation_index]
        vessels_in_operation = this_operation.vessels
        last_vessel_to_enter_lock = vessels_in_operation[-1] == vessel

        doors_can_be_closed = False
        if not between_arrivals and not self.lock_master.closing_doors_in_between_operations:
            return doors_can_be_closed
        if between_arrivals and (not self.closing_doors_in_between_arrivals or not last_vessel_to_enter_lock):
            return doors_can_be_closed
        doors_can_be_closed = True

        operation_planning = self.lock_master.operation_planning
        vessel_planning = self.lock_master.vessel_planning

        if not between_arrivals:
            last_time_doors_closed = operation_planning.loc[operation_index, "time_potential_lock_door_closure_start"]
        else:
            last_time_doors_closed = pd.Timestamp(datetime.datetime.fromtimestamp(self.env.now))
        last_time_doors_closed += pd.Timedelta(seconds=self.doors_closing_time)

        next_operations = operation_planning[operation_planning.index > operation_index]
        vessel_index = operation_planning.loc[operation_index, "vessels"].index(vessel)
        vessels_in_operation = operation_planning.loc[operation_index, "vessels"]

        operation_step = 1
        if between_arrivals and vessel_index != len(vessels_in_operation) - 1:
            next_vessel = vessels_in_operation[vessel_index + 1]
            next_vessel_planning_index = vessel_planning[vessel_planning.id == next_vessel.id].iloc[-1].name
            doors_required_to_be_open = vessel_planning.loc[next_vessel_planning_index, "time_potential_lock_door_opening_stop"]
            same_direction = True
        elif not next_operations.empty:
            next_operation = next_operations.iloc[0]
            if not len(next_operation.vessels):
                next_operation = next_operations.iloc[1]
                operation_step += 1
            doors_required_to_be_open = next_operation.time_potential_lock_door_opening_stop
            same_direction = direction != next_operation.direction
        else:
            return doors_can_be_closed

        if same_direction:
            direction = 1 - direction

        door_opening_time = self.determine_time_to_open_door(operation_index + operation_step, direction, doors_required_to_be_open)

        if (
            doors_required_to_be_open - door_opening_time < last_time_doors_closed
            or doors_required_to_be_open - last_time_doors_closed
            < self.minimum_time_between_operations_for_intermediate_door_closure
        ):
            doors_can_be_closed = False
        return doors_can_be_closed

    def determine_if_door_is_closed(self, vessel, operation_index, direction, first_in_lock=False, between_arrivals=False):
        """
        Determines if the doors are closed

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        operation_index : int
            index of the lock operation
        direction : int
            the direction of the lock operation: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        first_in_lock : bool
            if the function is run for the first vessel assigned to the lock operation (True) or not (False)
        between_arrivals : bool
            if the function is run to determine if the doors can be closed in between vessel arrivals (True) or not (False)

        Returns
        -------
        doors_are_closed : bool
            doors are closed (True) or not (False)
        doors_required_to_be_open : pd.Timestamp
            moment in time when the doors need to be opened
        operation_time : pd.Timedelta
            the time duration required to perform the lock operation
        """
        operation_planning = self.lock_master.operation_planning
        vessel_planning = self.lock_master.vessel_planning
        vessels = operation_planning.loc[operation_index, "vessels"]
        vessel_index = vessels.index(vessel)

        if between_arrivals and not self.closing_doors_in_between_arrivals:
            return False, None, None

        if not between_arrivals and not self.lock_master.closing_doors_in_between_operations:
            return False, None, None

        last_lockage_was_empty = False
        if operation_index - 2 in operation_planning.index:
            last_lockage_was_empty = len(operation_planning.loc[operation_index - 1, "vessels"]) == 0
        if last_lockage_was_empty:
            return False, None, None

        if not first_in_lock and vessel_index:
            previous_vessel_planning_index = (
                vessel_planning[vessel_planning.id == operation_planning.loc[operation_index, "vessels"][vessel_index - 1].id]
                .iloc[-1]
                .name
            )
            last_time_doors_closed = vessel_planning.loc[
                previous_vessel_planning_index, "time_potential_lock_door_closure_start"
            ] + pd.Timedelta(seconds=self.doors_closing_time)
        elif operation_index == 0:
            last_time_doors_closed = datetime.datetime.fromtimestamp(self.env.now)
        else:
            last_time_doors_closed = operation_planning.loc[
                operation_index - 1
            ].time_potential_lock_door_closure_start + pd.Timedelta(seconds=self.doors_closing_time)

        if first_in_lock:
            doors_required_to_be_open = operation_planning.loc[operation_index, "time_potential_lock_door_opening_stop"]
        else:
            vessel_planning_index = vessel_planning[vessel_planning.id == vessel.id].iloc[-1].name
            doors_required_to_be_open = vessel_planning.loc[vessel_planning_index, "time_potential_lock_door_opening_stop"]

        operation_time = self.determine_time_to_open_door(operation_index, direction, doors_required_to_be_open)
        doors_are_closed = False

        if (
            doors_required_to_be_open - operation_time > last_time_doors_closed
            and doors_required_to_be_open - last_time_doors_closed
            > self.minimum_time_between_operations_for_intermediate_door_closure
        ):
            doors_are_closed = True

        return doors_are_closed, doors_required_to_be_open, operation_time

    def determine_time_to_open_door(self, operation_index, direction, doors_required_to_be_open):
        """
        Determines the time to finish the levelling process and the door opening process

        Parameters
        ----------
        operation_index : int
            index of the lock operation
        direction : int
            the direction of the lock operation: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        doors_required_to_be_open : pd.Timestamp
            the moment in time that the doors are required to be opened

        Returns
        -------
        operation_time : pd.Timedelta
            the time to finish the levelling process and the door opening process
        """
        last_entering_time = doors_required_to_be_open - pd.Timedelta(seconds=self.doors_opening_time)
        operation_start_time = doors_required_to_be_open - pd.Timedelta(seconds=self.doors_opening_time)
        levelling_information = self.calculate_lock_operation_times(
            operation_index=operation_index,
            last_entering_time=last_entering_time,
            start_time=operation_start_time,
            direction=direction,
        )

        levelling_time = levelling_information["time_levelling_stop"] - levelling_information["time_levelling_start"]
        wlev_before, wlev_after = levelling_information["wlev_A"], levelling_information["wlev_B"]

        levelling_required = True
        if abs(wlev_after - wlev_before) < 0.1:
            levelling_required = False

        if not levelling_required:
            levelling_time = pd.Timedelta(seconds=0.0)

        operation_time = levelling_time + pd.Timedelta(seconds=self.doors_opening_time)
        return operation_time

    def determine_water_levels_before_and_after_levelling(self, levelling_start, levelling_stop, direction):
        """
        Determines the water level at both sides of the lock

        Parameters
        ----------
        levelling_start : pd.Timestamp
            the start time of the levelling process
        levelling_stop : pd.Timestamp
            the stop time of the levelling process
        direction : int
            the direction of the lock operation: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)

        Returns
        ------
        wlev_A :
            the water level at side A [m] before or after the levelling process (depending on the direction of the operation)
        wlev_B :
            the actual water level at side B [m] before or after the levelling process (depending on the direction of the operation)

        """
        hydromanager = HydrodynamicDataManager()
        t_start = np.datetime64(levelling_start)
        t_stop = np.datetime64(levelling_stop)
        if not direction:
            wlev_A = hydromanager._get_hydrodynamic_data_value(
                self.env.vessel_traffic_service.hydrodynamic_information_path, t_start, self.start_node, "Water level"
            )
            wlev_B = hydromanager._get_hydrodynamic_data_value(
                self.env.vessel_traffic_service.hydrodynamic_information_path, t_stop, self.end_node, "Water level"
            )
        else:
            wlev_A = hydromanager._get_hydrodynamic_data_value(
                self.env.vessel_traffic_service.hydrodynamic_information_path, t_stop, self.start_node, "Water level"
            )
            wlev_B = hydromanager._get_hydrodynamic_data_value(
                self.env.vessel_traffic_service.hydrodynamic_information_path, t_start, self.end_node, "Water level"
            )

        return wlev_A, wlev_B

    def determine_levelling_time(self, t_start, direction, wlev_init=None, operation_index=0, prediction=False):
        """
        Calculates the levelling time of a lock operation

        Parameters
        ----------
        t_start :
            the start time of the levelling process
        direction : int
            the direction of the vessel: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        wlev_init : float
            initial water level in the lock chamber
        same_direction : bool
            states if the levelling process is predicted in the same direction as the last lock operation (True) or not (False)
        prediction : bool
            states if the levelling process is only predicted (True) or executed (False)

        Returns
        -------
        levelling_time : float
            the time duration of the levelling process
        t : list of float
            the time series of the levelling process
        z : list of float
            the water level difference series over the time of the levelling process
        """
        # TODO: functie maken om tstart om te zetten (met _to_array)
        # TODO: Bij andere klasses altijd checken of iets een datetime.datetime is. En als dit niet zo is een error inbouwen of hem gelijk omzetten.
        dt = self.time_step
        t_final = 3600  # maximum levelling time has been set to an hour
        t = np.arange(0, t_final + float(dt), float(dt))
        t_start = time_to_numpy(t_start)
        # if there is no hydrodynamic data included in the run, use the constant levelling time included in the lock object
        # if there is no hydrodynamic data included in the run, use the constant levelling time included in the lock object
        if self.env.vessel_traffic_service.hydrodynamic_information_path is None:
            levelling_time = self.levelling_time
            z = np.zeros(len(t))
            return levelling_time, t, z

        z, H_A, H_B = calculate_z(
            t=t,
            t_start=t_start,
            direction=direction,
            wlev_init=wlev_init,
            operation_index=operation_index,
            operation_planning=self.lock_master.operation_planning,
            hydrodynamic_information_path=self.env.vessel_traffic_service.hydrodynamic_information_path,
            start_node=self.start_node,
            end_node=self.end_node,
            node_open=self.node_open,
            epoch=self.env.epoch,
        )

        # if a function has been included to predict the levelling time based on the water level difference: calculate the levelling time based on the initial water level difference
        if callable(self.levelling_time):
            levelling_time = self.levelling_time(z[0])
            return levelling_time, t, z

        # if no function has been included: compute the levelling time based on Eq. 4.64 of Ports and Waterways Open Textbook (https://books.open.tudelft.nl/home/catalog/book/204)
        levelling_time, t, z = levelling_time_equation(
            t=t,
            z=z,
            lock_length=self.lock_length,
            lock_width=self.lock_width,
            disch_coeff=self.disch_coeff,
            gate_opening_time=self.gate_opening_time,
            opening_area=self.opening_area,
            t_start=t_start,
            dt=dt,
            direction=direction,
            water_level_difference_limit_to_open_doors=self.lock_master.water_level_difference_limit_to_open_doors,
            prediction=prediction,
            H_A=H_A,
            H_B=H_B,
        )

        # if this function was not ran as a prediction, but rather as the actual levelling event: update the water level time series of the lock chamber
        if not prediction:
            # TODO: de self.water_level wordt niet gebruikt, maar is wel leuk om als logging terug te zien na een berekening. Nadenken of we dat zo willen laten, of anders willen bijhouden.
            t_final = t_start + np.timedelta64(int(levelling_time))
            t_index_final = HydrodynamicDataManager()._get_time_index_of_hydrodynamic_data(
                self.env.vessel_traffic_service.hydrodynamic_information_path, t_final
            )
            if not direction:
                self.water_level[t_index_final:] = H_B[t_index_final:].copy()
            else:
                self.water_level[t_index_final:] = H_A[t_index_final:].copy()

        return levelling_time, t, z


class IsLockChamber(IsLockChamberOperator, HasResource, HasLength, Identifiable, Log, HasOutput, HasMultiDiGraph, ExtraMetadata):
    """Mixin class: lock complex has a lock chamber:

    creates a lock chamber with a resource which is requested when a vessels wants to enter the area with limited capacity

    Attributes
    ----------
    vessel_sailing_speed_in_lock :
        calculates the average speed in the lock when entering
    vessel_sailing_speed_out_lock :
        calculates the average speed in the lock when leaving
    vessel_sailing_in_speed :
        Calculates the average speed when sailing towards the lock chamber
    vessel_sailing_out_speed :
        Calculates the average speed when sailing away from the lock chamber
    determine_levelling_time :
        calculates the levelling time of a lock operation

    """

    def __init__(
        self,
        start_node,  # a string which indicates the location of the first pair of lock doors
        end_node,  # a string which indicates the location of the second pair of lock doors
        lock_length,  # a float which contains the length of the lock chamber
        lock_width,  # a float which contains the width of the lock chamber
        lock_depth,  # a float which contains the depth of the lock chamber
        lock_master=None,
        k=0,  # a int which is the identifier of the edge between two nodes in a multidigraph network
        distance_from_start_node_to_lock_doors_A=0.0,  # a float that is the distance between the start_node of the edge and the lock doors A [m]
        distance_from_end_node_to_lock_doors_B=0.0,  # a float that is the distance between the end_node of the edge and the lock doors B [m]
        registration_nodes=[],  # a list of str with the node names at which the vessels request registration to the lock complex master
        doors_opening_time=300.0,  # a float which contains the time it takes to open the doors [s]
        doors_closing_time=300.0,  # a float which contains the time it takes to close the doors [s]
        disch_coeff=0.4,  # a float which contains the discharge coefficient of filling system
        opening_area=12.0,  # a float which contains the cross-sectional area of filling system [m^2]
        opening_depth=None,  # a float which contains the depth at which filling system is located [m^2]
        speed_reduction_factor_lock_chamber=0.3,  # a float that is the reduction factor for the vessel speed from its original speed when entering the lock
        start_sailing_out_time_after_doors_have_been_opened=0.0,  # a float that is the time that the vessel wait to start sailing out of the lock after the doors have been opened after levelling [s]
        sailing_time_before_opening_lock_doors=600.0,  # a float that is the time that the doors are opened before a vessel arrives at the doors [s]
        sailing_time_before_closing_lock_doors=120.0,  # a float that is the time that the doors are closed after a vessel has sailed through the doors [s]
        minimum_time_between_operations_for_intermediate_door_closure=0.0,  # a float that is the minimum required time between lock operations that the lock doors can be both closed (to reduce salt intrusion) [s]
        sailing_distance_to_crossing_point=500.0,  # a float that is the distance at which vessels can safely pass each other in front of the lock (last vessel that sails out and first vessel that sails in) [m]
        passage_time_door=300.0,  # a float [s] ?
        sailing_in_time_gap_through_doors=180.0,  # a float that is the time gap after which the next vessel can sail into the lock through the lock doors (after another vessel has sailed through to enter the lock) [s]
        sailing_out_time_gap_through_doors=180.0,  # a float that is the time gap after which the next vessel can sail out of the lock through the lock doors (after another vessel has sailed through to leave the lock)[s]
        sailing_in_time_gap_after_berthing_previous_vessel=0.0,  # a float that is the time gap after which the next vessel can sail into the lock (after another vessel has berthed) [s]
        sailing_out_time_gap_after_berthing_previous_vessel=0.0,  # a float that is the time gap after which the next vessel can sail out of the lock (after another vessel has deberthed) [s]
        sailing_in_speed_A=2 * knots,  # a float that is the speed at which the vessel sails into the lock to the sea side [m/s]
        sailing_out_speed_A=2 * knots,  # a float that is the speed at which the vessel sails out of the lock to the sea side [m/s]
        sailing_in_speed_B=2 * knots,  # a float that is the speed at which the vessel sails into the lock to the canal side [m/s]
        sailing_out_speed_B=2
        * knots,  # a float that is the speed at which the vessel sails out of the lock to the canal side [m/s]
        minimum_manoeuvrability_speed=2
        * knots,  # a float that is the minimum speed at which the vessel is still safely manoeuvrable [m/s]
        levelling_time=600.0,  # a float that fixates the levelling time [s]
        time_step=10.0,  # a float that is the integration time step to determine the levelling time [s]
        gate_opening_time=60.0,  # a float that is the time it takes for the levelling gate to open [s]
        node_open=None,  # a string that is the node name to which the lock was last levelled to at the initial time of simulation (either start_node or end_node)
        conditions=None,  # maybe obsolete ???
        priority_rules=None,  # maybe obsolete ???
        used_as_one_way_traffic_regulation=False,  # maybe obsolete ???
        seed_nr=None,  # a int for the seed to fix the determination of the node_open when node_open is None
        *args,
        **kwargs,
    ):
        """Initialization"""
        # TODO: checken of alle inputs nodig zijn
        # TODO: checken of alle parents nodig zijn
        # TODO: parentklasse Lockmaster toevoegen

        # set input parameters as properties
        self.lock_length = lock_length
        self.lock_width = lock_width
        # TODO: @Floor lock_depth wordt niet gebruikt... Willen we die houden?
        self.lock_depth = lock_depth
        # TODO @Floor, is deze coefficient afhankelijk van de lock, of is dit een standaard coefficient die we ergens anders kunnen opslaan?
        self.disch_coeff = disch_coeff #0.4

        self.opening_area = opening_area
        if opening_depth is None:
            opening_depth = lock_depth/2
        self.opening_depth = opening_depth
        self.levelling_time = levelling_time
        self.start_sailing_out_time_after_doors_have_been_opened = start_sailing_out_time_after_doors_have_been_opened
        self.sailing_time_before_opening_lock_doors = sailing_time_before_opening_lock_doors
        self.sailing_time_before_closing_lock_doors = sailing_time_before_closing_lock_doors
        self.minimum_time_between_operations_for_intermediate_door_closure = minimum_time_between_operations_for_intermediate_door_closure
        self.sailing_in_time_gap_after_berthing_previous_vessel = sailing_in_time_gap_after_berthing_previous_vessel
        self.sailing_out_time_gap_after_berthing_previous_vessel = sailing_out_time_gap_after_berthing_previous_vessel
        self.sailing_in_speed_A = sailing_in_speed_A
        self.sailing_out_speed_A = sailing_out_speed_A
        self.sailing_in_speed_B = sailing_in_speed_B
        self.sailing_out_speed_B = sailing_out_speed_B
        self.sailing_distance_to_crossing_point = sailing_distance_to_crossing_point
        self.sailing_in_time_gap_through_doors = sailing_in_time_gap_through_doors
        self.sailing_out_time_gap_through_doors = sailing_out_time_gap_through_doors
        self.speed_reduction_factor = speed_reduction_factor_lock_chamber
        self.passage_time_door = passage_time_door
        self.start_node = start_node
        self.end_node = end_node
        self.k = k
        self.minimum_manoeuvrability_speed = minimum_manoeuvrability_speed
        self.node_open = node_open
        self.conditions = conditions
        self.time_step = time_step
        self.priority_rules = priority_rules
        self.registration_nodes = registration_nodes
        self.gate_opening_time = gate_opening_time
        self.door_A_open = True
        self.door_B_open = True
        if not registration_nodes:
            self.registration_nodes = [start_node,end_node]
        self.distance_from_start_node_to_lock_doors_A = distance_from_start_node_to_lock_doors_A
        self.distance_from_end_node_to_lock_doors_B = distance_from_end_node_to_lock_doors_B
        self.used_as_one_way_traffic_regulation = used_as_one_way_traffic_regulation
        self.converting_chamber = False

        # TODO: checken of de seed_nr en de random functie worden gebruikt.
        if seed_nr is not None:
            np.random.seed(seed_nr)

        # TODO: als lockmaster een parent klasse is, zou lock_complex=self weg moeten kunnen.
        # TODO: capaciteit = 100. checken of deze info overbodig is doordat er al een lock_length is. En anders kijken of de capaciteit op oneindig kan.
        super().__init__(
            lock_master=lock_master,
            capacity=100,
            length=lock_length,
            remaining_length=lock_length,
            *args,
            **kwargs,
        )

        self._verify_node_AB()

        if self.env.vessel_traffic_service.hydrodynamic_information_path is not None:
            hydro_manager = HydrodynamicDataManager()
            if isinstance(self.env.vessel_traffic_service.hydrodynamic_information_path,str):
                hydro_manager.hydrodynamic_data = Dataset(self.env.vessel_traffic_service.hydrodynamic_information_path)
            else:
                hydro_manager.hydrodynamic_data = self.env.vessel_traffic_service.hydrodynamic_information
            if isinstance(self.env.vessel_traffic_service.hydrodynamic_information_path, str):
                hydro_manager.hydrodynamic_times = (
                    hydro_manager.hydrodynamic_data["TIME"][:].data.astype("timedelta64[m]")
                    + self.env.vessel_traffic_service.hydrodynamic_start_time
                )
            else:
                hydro_manager.hydrodynamic_times = hydro_manager.hydrodynamic_data["TIME"][:]

        if self.node_open is None:
            self.node_open = np.random.choice([start_node, end_node])

        if self.closing_doors_in_between_operations:
            self.door_A_open = False
            self.door_B_open = False
        elif self.node_open == self.start_node:
            self.door_B_open = False
        else:
            self.door_A_open = False

        # Geometry on edge
        edge = (start_node, end_node, 0)
        edge_info = self.multidigraph.edges[edge]
        # TODO Checken of de distance bepalen werkt, en misschien automatiseren op basis van geometrie
        # TODO: nodes verwijderen uit graaf als die precies op de sluis liggen. (wellicht als voorbewerking van de graaf)
        # TODO: losse klasse maken van de lock-doors die locatable, hasresource (capacity=1) en identifiable is en eigenschap open/dicht heeft.

        edge_aligned_with_edge_geometry = self.env.vessel_traffic_service.check_if_geometry_is_aligned_with_edge(edge)
        start_node_geometry = start_node
        end_node_geometry = end_node
        distance_from_start_node_geometry_to_lock_doors_A = self.distance_from_start_node_to_lock_doors_A
        distance_from_start_node_geometry_to_lock_doors_B = self.distance_from_start_node_to_lock_doors_A + lock_length
        if not edge_aligned_with_edge_geometry:
            start_node_geometry = end_node
            end_node_geometry = start_node
            distance_from_start_node_geometry_to_lock_doors_B = self.distance_from_end_node_to_lock_doors_B
            distance_from_start_node_geometry_to_lock_doors_A = self.distance_from_end_node_to_lock_doors_B + lock_length

        self.location_lock_doors_A = self.env.vessel_traffic_service.provide_location_over_edges(start_node_geometry, end_node_geometry, distance_from_start_node_geometry_to_lock_doors_A)
        self.location_lock_doors_B = self.env.vessel_traffic_service.provide_location_over_edges(start_node_geometry, end_node_geometry, distance_from_start_node_geometry_to_lock_doors_B)

        self.lock_pos_length = simpy.Container(self.env, capacity=lock_length, init=lock_length)
        self.door_A= simpy.PriorityResource(self.env, capacity = 1)
        self.levelling = simpy.Resource(self.env, capacity=1)
        self.door_B = simpy.PriorityResource(self.env, capacity = 1)

        # TODO: kijken of onderstaande eigenschappen nodig zijn. en capacity op infinity zetten als mogelijk.
        self.wait_for_other_vessel_to_arrive = simpy.FilterStore(self.env,capacity=100000000)
        self.wait_for_levelling = simpy.FilterStore(self.env,capacity=100000000)
        self.wait_for_other_vessels = simpy.FilterStore(self.env,capacity=100000000)

        # Operating
        self.doors_opening_time = doors_opening_time
        self.doors_closing_time = doors_closing_time

        # TODO: maak ene functie _test_input() die bijvoorbeeld checkt of de nodes na elkaar liggen.
        # Water level
        assert start_node != end_node

        time = np.datetime64(datetime.datetime.fromtimestamp(self.env.now))
        wlev_series = HydrodynamicDataManager()._get_hydrodynamic_data_series(
            self.env.vessel_traffic_service.hydrodynamic_information_path, time, self.node_open, "Water level"
        )
        self.water_level = wlev_series

        # TODO: in functie zetten.
        # TODO: In de documentatie zetten dat detecotr nodes op volgorde moeten komen. En ook een assert maken.
        for registration_node, lock_edge in zip(self.registration_nodes,[(self.start_node,self.end_node,self.k),(self.end_node,self.start_node,self.k)]):
            if 'Lock_registration_node' not in self.multidigraph.nodes[registration_node]:
                self.multidigraph.nodes[registration_node]['Lock_registration_node'] = lock_edge

        # Add to the graph:
        # TODO: In losse functie (add_lock_to_graph)
        if "graph" in dir(self.env):
            k = sorted(self.multidigraph[self.start_node][self.end_node], key=lambda x: get_length_of_edge(self.multidigraph,(self.start_node, self.end_node, x)))[0]
            # Add the lock to the edge or append it to the existing list
            if "Lock" not in self.multidigraph.edges[self.start_node, self.end_node, k].keys():
                self.multidigraph.edges[self.start_node, self.end_node, k]["Lock"] = [self]
                self.multidigraph.edges[self.end_node, self.start_node, k]["Lock"] = [self]
            else:
                self.multidigraph.edges[self.start_node, self.end_node, k]["Lock"].append(self)
                self.multidigraph.edges[self.end_node, self.start_node, k]["Lock"].append(self)

    def _verify_node_AB(self):
        """Function to verify if nodes A and B are part of the graph, and have an edge between them."""
        if self.start_node not in self.env.graph.nodes or self.end_node not in self.env.graph.nodes:
            raise ValueError(
                f"Lock chamber {self.name} has invalid node_A {self.start_node} or node_B {self.end_node} which are not part of the graph."
            )
        if not self.env.graph.has_edge(self.start_node, self.end_node):
            raise ValueError(
                f"Lock chamber {self.name} does not have an edge between node A {self.start_node} and node B {self.end_node}."
            )

    def vessel_sailing_speed_in_lock(self, vessel):
        """
        Calculates the average speed in the lock when entering

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        x_location_lock : float
            logintudinal coordinate in the lock to which the vessel is assigned [m]
        P_used : float
            the breaking power used by the vessel to gradually decelerate [kW]

        Returns
        -------
        speed : float
            the average speed in the lock from the lock doors to the location of berthing

        """
        # TODO: sailing_in_speed_B zou A of B moeten zijn. Checken of deze eigenschap vaker voorkomt.
        speed = self.sailing_in_speed_B
        if vessel.bound == 'inbound':
            speed = self.sailing_in_speed_A

        return speed

    def vessel_sailing_speed_out_lock(self, vessel):
        """
        Calculates the average speed to in the lock when leaving

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        x_location_lock : float
            logintudinal coordinate in the lock to which the vessel is assigned [m]
        P_used : float
            the breaking power used by the vessel to gradually decelerate [kW]

        Returns
        -------
        speed : float
            the average speed in the lock from the lock doors to the location of berthing

        """
        speed = self.sailing_out_speed_A
        if vessel.bound == 'inbound':
            speed = self.sailing_out_speed_B

        return speed

    def vessel_sailing_in_speed(self, vessel, direction):
        """
        Calculates the average speed when sailing towards the lock chamber

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        x_location_lock : float
            logintudinal coordinate in the lock to which the vessel is assigned [m]
        P_used : float
            the breaking power used by the vessel to gradually decelerate [kW]

        Returns
        -------
        speed : float
            the average speed in the lock from the lock doors to the location of berthing

        """
        # determine the edge on which the vessel is sailing and the distance to the lock doors
        edge = self._directional_edge(direction)

        # determine the speed of the vessel over the edge
        speed = vessel._compute_velocity_on_edge(edge[0], edge[1])

        # if there is an overruled speed on the edge, use this speed
        if "overruled_speed" in dir(vessel) and edge in vessel.overruled_speed.index:
            speed = vessel.overruled_speed.loc[edge, "Speed"]

        return speed

    def vessel_sailing_out_speed(self, vessel, direction, P_used=None, h0=17, until_crossing_point=False):
        """
        Calculates the average speed when sailing away from the lock chamber

        Parameters
        ----------
        vessel : type
            a type including the following parent-classes: PassesLockComplex, Identifiable, Movable, VesselProperties, ExtraMetadata, HasMultiDiGraph, HasOutput
        direction : int
            the direction of the vessel: 0 (direction from node_A to node_B) or 1 (direction from node_B to node_A)
        P_used : float
            the breaking power used by the vessel to gradually decelerate [kW]
        until_crossing_point : bool


        Returns
        -------
        speed : float
            the average speed in the lock from the lock doors to the location of berthing

        """
        # determine the edge on which the vessel is sailing and the distance to the lock doors
        edge = self._directional_edge(direction)

        # determine the speed of the vessel over the edge
        speed = vessel._compute_velocity_on_edge(edge[0], edge[1])

        # if there is an overruled speed on the edge, use this speed
        if 'overruled_speed' in dir(vessel) and edge in vessel.overruled_speed.index:
            speed = vessel.overruled_speed.loc[edge, 'Speed']

        return speed

    def _directional_edge(self, direction):
        """get the edge of the lock chamber in the correct direction"""
        if not direction:
            return (self.start_node, self.end_node)
        else:
            return (self.end_node, self.start_node)



from .mixins import (
    load_partial_engine_load_correction_factors,
    karpov_smooth_curves,
    ConsumesEnergy,
    EnergyCalculation,
)
from . import logutils


__all__ = [
    "load_partial_engine_load_correction_factors",
    "karpov_smooth_curves",
    "ConsumesEnergy",
    "EnergyCalculation",
    "logutils",
]


"""
Mixin classes related to resources.

The following classes are provided:
- HasResource

"""
# package(s) for documentation, debugging, saving and loading
import logging

# package(s) related to the simulation
import simpy

# use OpenCLSim objects for core objects (identifiable is imported for later use)
from openclsim.core import SimpyObject

# get logger
logger = logging.getLogger(__name__)


class HasResource(SimpyObject):
    """Mixin class: Something that has a resource limitation, a resource request must be granted before the object can be used.

    Parameters
    -----------
    nr_resources: int, default=1
        nr of requests that can be handled simultaneously, optional, default=1
    priority: bool, default=False
        if True, prioritized resources can be handled. optional, default=False.
    args, kwargs:
        passed to SimpyObject. Must at least contain parameter env: simpy.Environment.

    Attributes
    -----------
    resource: simpy.Resource or simpy.PriorityResource
        the resource that is used to limit the nr of requests that can be handled simultaneously.
    env: simpy.Environment
        the simpy environment that is used to run the simulation.
    """

    def __init__(self, nr_resources: int = 1, priority: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """Initialization"""
        self.resource = (
            simpy.PriorityResource(self.env, capacity=nr_resources) if priority else simpy.Resource(self.env, capacity=nr_resources)
        )


"""Mixins for Energy

This script contains the classes
    - ConsumesEnergy: a mixin for objects that consume energy
    - EnergyCalculations: a class to perform energy calculations and save results in a logbook. This is not a mixin, but must be called after the simulation.
"""

# %% IMPORT DEPENENDENCIES
# generic
import pathlib
import logging
import numpy as np
import pandas as pd

# OpenTNSim
import opentnsim
from opentnsim.graph.mixins import calculate_distance, calculate_depth
from opentnsim.energy.algorithms import power2v
from opentnsim.energy.calculations import (
    sample_engine_age,
    calculate_properties,
    calculate_frictional_resistance,
    calculate_viscous_resistance,
    calculate_appendage_resistance,
    karpov,
    calculate_wave_resistance,
    calculate_residual_resistance,
    calculate_total_power_required,
    calculate_max_sinkage,
    calculate_total_resistance,
)


# logging
logger = logging.getLogger(__name__)


# %% AUXILIARY FUNCTIONS
def load_partial_engine_load_correction_factors():
    """read correction factor from package directory"""

    # Can't get this  to work with pkg_resourcs
    data_dir = pathlib.Path(__file__).parent.parent / "data"
    correctionfactors_path = data_dir / "Correctionfactors.csv"
    df = pd.read_csv(correctionfactors_path, comment="#")

    return df


def karpov_smooth_curves():
    """read correction factor from package directory"""

    # Can't get this  to work with pkg_resourcs
    data_dir = pathlib.Path(__file__).parent.parent / "data"
    karpov_smooth_curves_path = data_dir / "KarpovSmoothCurves.csv"
    df = pd.read_csv(karpov_smooth_curves_path, comment="#")

    return df


# %% CLASSES


class ConsumesEnergy:
    """
    Mixin class: Something that consumes energy.

    Parameters
    ----------
    P_installed : float
         Installed engine power in kilowatts (kW).
    P_tot_given : float
        Total power set by captain (includes hotel power). When
        P_tot_given > P_installed; P_tot_given=P_installed.
    karpov_correction : bool, optional
        If True, apply Karpovs correction for velocity under the vessel, if False,
        use the speed to water.
    bulbous_bow : bool, optional
        Indicates if the ship has a bulbous bow. Inland ships generally do
        not have a bulbous bow, hence the default is False. If a ship has
        a bulbous bow, set to True.
    L_w : int
        Weight class of the ship depending on carrying capacity. Classes:
        L1 (=1), L2 (=2), L3 (=3).
    current_year : int
        The current year.
    nu : float
        Kinematic viscosity in square meters per second (m^2/s).
    rho : float
        Density of the surrounding water in kilograms per cubic meter
        (kg/m^3).
    g : float
        Gravitational acceleration in meters per second squared (m/s^2).
    x : int
        Number of propellers.
    eta_o : float
        Open water efficiency of the propeller.
    eta_r : float
        Relative rotative efficiency.
    eta_t : float
        Transmission efficiency.
    eta_g : float
        Gearing efficiency.
    c_stern : float
        Determines the shape of the afterbody.
    C_BB : float
        Breadth coefficient of the bulbous bow, set to 0.2 according to the
        paper of Kracht (1970), https://doi.org/10.5957/jsr.1970.14.1.1.
    C_B : float, optional
        Block coefficient ('fullness'), default to 0.85.
    one_k2 : float
        Appendage resistance factor (1+k2).
    C_year : int
        Construction year of the engine.
    """

    def __init__(
        self,
        P_installed,
        L_w,
        C_year=None,
        current_year=None,  # current_year
        bulbous_bow=False,
        karpov_correction=False,
        P_hotel_perc=0.05,
        P_hotel=None,
        P_tot_given=None,  # the actual power engine setting
        nu=1 * 10 ** (-6),
        rho=1000,
        g=9.81,
        x=2,
        D_s=1.4,
        eta_o=0.4,
        eta_r=1.00,
        eta_t=0.98,
        eta_g=0.96,
        c_stern=0,
        C_BB=0.2,
        C_B=0.85,
        one_k2=2.5,  # following Segers (2021) we assume (1 + k2) to be 2.5 (see below Eq 3.27)
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        """Initialization
        """

        self.P_installed = P_installed
        self.bulbous_bow = bulbous_bow
        self.karpov_correction = karpov_correction

        # Required power for systems on board, "5%" based on De Vos and van Gils (2011): Walstroom versus generator stroom
        self.P_hotel_perc = P_hotel_perc

        if P_hotel:  # if P_hotel is specified use the given value
            self.P_hotel = P_hotel
        else:  # if P_hotel is None calculate it from P_hotel_percentage and P_installed
            self.P_hotel = self.P_hotel_perc * self.P_installed

        self.P_tot_given = P_tot_given
        self.L_w = L_w
        self.year = current_year
        self.nu = nu
        self.rho = rho
        self.g = g
        self.x = x
        self.D_s = D_s
        self.eta_o = eta_o
        self.eta_r = eta_r
        self.eta_t = eta_t
        self.eta_g = eta_g
        self.c_stern = c_stern
        self.C_BB = C_BB
        self.C_B = C_B

        self.one_k2 = one_k2

        # plugin function that computes velocity based on power
        self.power2v = opentnsim.energy.algorithms.power2v

        # TODO: C_year is obligatory, so why is this code here?
        if C_year:
            self.C_year = C_year
        else:
            self.C_year = self.calculate_engine_age()

        if self.P_tot_given is not None and self.P_installed is not None:
            if P_tot_given > P_installed:
                self.P_tot_given = self.P_installed

    def calculate_engine_age(self):
        """
        Calculate the age of the engine based on the weight class of the ship (L_w).

        The age is drawn randomly from a Weibull distribution with parameters k and lmb,
        which depend on the weight class of the ship. The year of construction is
        computed from the current year of the simulation and the age of the engine.

        Notes
        -----
        Uses `self.L_w` and `self.year` to compute the age and construction year of the
        engine. This method sets attributes `self.age` and `self.C_year`.

        """
        # compute the age of the engine
        self.age = sample_engine_age(self.L_w)

        # compute the construction year of the engine
        if self.year is None:
            raise ValueError(
                "year must be set to calculate the construction year of the engine"
            )
        self.C_year = self.year - self.age
        logger.debug(
            f"Engine age calculated as {self.age}, hence construction year is {self.C_year}"
        )

        return self.C_year

    def calculate_properties(self):
        """Calculate a number of basic vessel properties"""

        (
            self.C_M,
            self.C_WP,
            self.C_P,
            self.delta,
            self.lcb,
            self.L_R,
            self.A_T,
            self.A_BT,
            self.S,
            self.S_APP,
            self.S_B,
            self.T_F,
            self.h_B,
        ) = calculate_properties(
            C_B=self.C_B,
            L=self.L,
            B=self.B,
            T=self.T,
            bulbous_bow=self.bulbous_bow,
            C_BB=self.C_BB,
        )

    def calculate_frictional_resistance(self, v, h_0):
        """Frictional resistance

        - 1st resistance component defined by Holtrop and Mennen (1982)
        - A modification to the original friction line is applied, based on literature of Zeng (2018), to account for shallow water effects
        """
        (
            self.R_f,
            self.C_f,
            self.R_e,
            self.Cf_deep,
            self.Cf_shallow,
            self.Cf_0,
            self.Cf_Katsui,
            self.V_B,
            self.D,
            self.a,
        ) = calculate_frictional_resistance(
            v=v,
            h_0=h_0,
            L=self.L,
            nu=self.nu,
            T=self.T,
            S=self.S,
            S_B=self.S_B,
            rho=self.rho,
        )

    def calculate_viscous_resistance(self):
        """Viscous resistance

        - 2nd resistance component defined by Holtrop and Mennen (1982)
        - Form factor (1 + k1) has to be multiplied by the frictional resistance R_f, to account for the effect of viscosity
        """

        self.c_14, self.one_k1, self.R_f_one_k1 = calculate_viscous_resistance(
            c_stern=self.c_stern,
            B=self.B,
            L=self.L,
            T=self.T,
            L_R=self.L_R,
            C_P=self.C_P,
            R_f=self.R_f,
            delta=self.delta,
        )

    def calculate_appendage_resistance(self, v):
        """Appendage resistance

        - 3rd resistance component defined by Holtrop and Mennen (1982)
        - Appendages (like a rudder, shafts, skeg) result in additional frictional resistance
        """

        self.R_APP = calculate_appendage_resistance(
            v=v,
            rho=self.rho,
            S_APP=self.S_APP,
            one_k2=self.one_k2,
            C_f=self.C_f,
        )

    def karpov(self, v, h_0):
        """Intermediate calculation: Karpov

        - The Karpov method computes a velocity correction that accounts for limited water depth (corrected velocity V2,
          expressed as "Vs + delta_V" in the paper), but it also can be used for deeper water depth (h_0 / T >= 9.5).
        - V2 has to be implemented in the wave resistance (R_W) and the residual resistance terms (R_res: R_TR, R_A, R_B)
        """

        self.F_rh, self.V_2, self.alpha_xx = karpov(
            v=v,
            h_0=h_0,
            g=self.g,
            T=self.T,
        )

    def calculate_wave_resistance(self, V_2, h_0):
        """Wave resistance

        - 4th resistance component defined by Holtrop and Mennen (1982)
        - When the speed or the vessel size increases, the wave making resistance increases
        - In shallow water, the wave resistance shows an asymptotical behaviour by reaching the critical speed
        """

        # perform calculation of wave resistance
        (
            self.F_rL,
            self.i_E,
            self.c_1,
            self.c_2,
            self.c_5,
            self.c_7,
            self.c_15,
            self.c_16,
            self.lmbda,
            self.m_1,
            self.m_2,
            self.R_W,
        ) = calculate_wave_resistance(
            V_2=V_2,
            h_0=h_0,
            g=self.g,
            T=self.T,
            L=self.L,
            B=self.B,
            C_P=self.C_P,
            C_WP=self.C_WP,
            lcb=self.lcb,
            L_R=self.L_R,
            A_T=self.A_T,
            C_M=self.C_M,
            delta=self.delta,
            rho=self.rho,
        )

    def calculate_residual_resistance(self, V_2, h_0):
        """Residual resistance terms

        - Holtrop and Mennen (1982) defined three residual resistance terms:
        - 1) Resistance due to immersed transom (R_TR), Karpov corrected velocity V2 is used
        - 2) Resistance due to model-ship correlation (R_A), Karpov corrected velocity V2 is used
        - 3) Resistance due to the bulbous bow (R_B), Karpov corrected velocity V2 is used
        """

        # compute the residual resistance terms
        (
            self.F_nT,
            self.c_6,
            self.R_TR,
            self.c_4,
            self.c_2,
            self.C_A,
            self.R_A,
            self.F_ni,
            self.P_B,
            self.R_B,
            self.R_res,
        ) = calculate_residual_resistance(
            V_2=V_2,
            g=self.g,
            A_T=self.A_T,
            B=self.B,
            C_WP=self.C_WP,
            rho=self.rho,
            T=self.T,
            L=self.L,
            C_B=self.C_B,
            S=self.S,
            T_F=self.T_F,
            h_B=self.h_B,
            A_BT=self.A_BT,
            bulbous_bow=self.bulbous_bow,
        )

    def calculate_total_resistance(self, v, h_0):
        """Total resistance:

        The total resistance is the sum of all resistance components (Holtrop and Mennen, 1982)
        """
        # TODO: we doen nu hier alle stappen en combineren dan alles, waarom roepen we hier niet
        #  calculate_total_resistance uit de calculations.py aan?

        self.calculate_properties()
        self.calculate_frictional_resistance(v, h_0)
        self.calculate_viscous_resistance()
        self.calculate_appendage_resistance(v)

        self.karpov(v, h_0)
        # print('Original v = {} m/s, Karpov corrected V_2 = {} m/s'.format(v, self.V_2))
        if self.karpov_correction:
            self.calculate_wave_resistance(self.V_2, h_0)
            self.calculate_residual_resistance(self.V_2, h_0)
        else:
            self.calculate_wave_resistance(v, h_0)
            self.calculate_residual_resistance(v, h_0)

        # The total resistance R_tot [kN] = R_f * (1+k1) + R_APP + R_W + R_TR + R_A
        self.R_tot = (
            self.R_f * self.one_k1
            + self.R_APP
            + self.R_W
            + self.R_TR
            + self.R_A
            + self.R_B
        )

    def calculate_total_power_required(self, v, h_0):
        """Total required power:

        - The total required power is the sum of the power for systems on board (P_hotel) + power required for
          propulsion
        - The power required for propulsion depends on the calculated resistance

        Output:
        - P_propulsion: required power for propulsion, equals to P_d (Delivered Horse Power)
        - P_tot: required power for propulsion and hotelling
        - P_given: the power given by the engine to the ship (for propulsion and hotelling), which is the actual power
          the ship uses

        Note:
        In this version, we define the propulsion power as P_d (Delivered Horse Power) rather than P_b (Brake Horse
        Power). The reason we choose P_d as propulsion power is to prevent double use of the same power efficiencies.
        The details are
        1) The P_b calculation involves gearing efficiency and transmission efficiency already while P_d not.
        2) P_d is the power delivered to propellers.
        3) To estimate the renewable fuel use, we will involve "energy conversion efficiencies" later in the
           calculation.
        The 'energy conversion efficiencies' for renewable fuel powered vessels are commonly measured/given as a whole
        covering the engine power systems, includes different engine (such as fuel cell engine, battery engine, internal
        combustion engine, hybrid engine) efficiencies, and corresponding gearbox efficiencies, AC/DC converter
        efficiencies, excludes the efficiency items of propellers.
        Therefore, to align with the later use of "energy conversion efficiencies" for fuel use estimation and prevent
        double use of some power efficiencies such as gearing efficiency, here we choose P_d as propulsion power.
        """
        (
            self.P_e,
            self.dw,
            self.w,
            self.t,
            self.eta_h,
            self.P_d,
            self.P_b,
            self.P_propulsion,
            self.P_tot,
            self.P_given,
            self.P_partial,
        ) = calculate_total_power_required(
            v=v,
            h_0=h_0,
            R_tot=self.R_tot,
            F_rL=self.F_rL,
            x=self.x,
            C_B=self.C_B,
            delta=self.delta,
            D_s=self.D_s,
            eta_o=self.eta_o,
            eta_r=self.eta_r,
            eta_t=self.eta_t,
            eta_g=self.eta_g,
            P_hotel=self.P_hotel,
            P_installed=self.P_installed,
        )

        # return these three variables:
        # 1) self.P_propulsion, for the convience of validation.  (propulsion power and fuel used for propulsion),
        # 2) self.P_tot, know the required power, especially when it exceeds installed engine power while sailing shallower and faster
        # 3) self.P_given, the actual power the engine gives for "propulsion + hotel" within its capacity (means installed power). This varible is used for calculating delta_energy of each sailing time step.
        # TODO: return description does not match the docstring and comments

        return self.P_given

    def emission_factors_general(self):
        """General emission factors:

        This function computes general emission factors, based on construction year of the engine.
        - Based on literature TNO (2019)

        Please note: later on a correction factor has to be applied to get the total emission factor
        """

        # The general emission factors of CO2, PM10 and NOX are based on the construction year of the engine

        if self.C_year < 1974:
            self.EF_CO2 = 756
            self.EF_PM10 = 0.6
            self.EF_NOX = 10.8
        if 1975 <= self.C_year <= 1979:
            self.EF_CO2 = 730
            self.EF_PM10 = 0.6
            self.EF_NOX = 10.6
        if 1980 <= self.C_year <= 1984:
            self.EF_CO2 = 714
            self.EF_PM10 = 0.6
            self.EF_NOX = 10.4
        if 1985 <= self.C_year <= 1989:
            self.EF_CO2 = 698
            self.EF_PM10 = 0.5
            self.EF_NOX = 10.1
        if 1990 <= self.C_year <= 1994:
            self.EF_CO2 = 698
            self.EF_PM10 = 0.4
            self.EF_NOX = 10.1
        if 1995 <= self.C_year <= 2002:
            self.EF_CO2 = 650
            self.EF_PM10 = 0.3
            self.EF_NOX = 9.4
        if 2003 <= self.C_year <= 2007:
            self.EF_CO2 = 635
            self.EF_PM10 = 0.3
            self.EF_NOX = 9.2
        if 2008 <= self.C_year <= 2019:
            self.EF_CO2 = 635
            self.EF_PM10 = 0.2
            self.EF_NOX = 7
        if self.C_year > 2019:
            if self.L_w == 1:
                self.EF_CO2 = 650
                self.EF_PM10 = 0.1
                self.EF_NOX = 2.1
            else:
                self.EF_CO2 = 603
                self.EF_PM10 = 0.015
                self.EF_NOX = 1.8

        logger.debug(f"The general emission factor of CO2 is {self.EF_CO2} g/kWh")
        logger.debug(f"The general emission factor of PM10 is {self.EF_PM10} g/kWh")
        logger.debug(f"The general emission factor CO2 is {self.EF_NOX} g/kWh")

    def energy_density(self):
        """net energy density of diesel and renewable energy sources. This will be used for calculating SFC later.

        - Edens_xx_mass: net gravimetric energy density, which is the amount of energy stored in a given energy source in mass [kWh/kg].
        - Edens_xx_vol: net volumetric energy density, which is the amount of energy stored in a given energy source in volume [kWh/m3].


        Data source:
        Table 3-2 from Marin report 2019,  Energietransitie emissieloze binnenvaart, vooronderzoek ontwerpaspecten,
        systeem configuraties.(Energy transition zero-emission inland shipping, preliminary research on design aspects,
        system configurations

        Note:
        net energy density can be used for calculate fuel consumption in mass and volume, but for required energy
        source storage space determination, the packaging factors of different energy sources also need to be
        considered.
        """

        # gravimetric net energy density
        self.Edens_diesel_mass = 11.67 / 1000  # kWh/kg
        self.Edens_LH2_mass = 33.3 / 1000  # kWh/kg
        self.Edens_eLNG_mass = 13.3 / 1000  # kWh/kg
        self.Edens_eMethanol_mass = 5.47 / 1000  # kWh/kg
        self.Edens_eNH3_mass = 5.11 / 1000  # kWh/kg
        self.Edens_Li_NMC_Battery_mass = 0.11 / 1000  # kWh/kg

        # volumetric net energy density
        self.Edens_diesel_vol = 9944  # kWh/m3
        self.Edens_LH2_vol = 2556  # kWh/m3
        self.Edens_eLNG_vol = 5639  # kWh/m3
        self.Edens_eMethanol_vol = 4333  # kWh/m3
        self.Edens_eNH3_vol = 3139  # kWh/m3
        self.Edens_Li_NMC_Battery_vol = 139  # kWh/m3

    def energy_conversion_efficiency(self):
        """energy efficiencies for combinations of different energy source and energy-power conversion systems, including engine and power plant, excluding propellers. This will be used for calculating SFC later.

        - Eeff_FuelCell: the efficiency of the fuel cell energy conversion system on board, includes fuel cells, AC/DC converter, electric motor and gearbox. Generally this value is between 40% - 60%, here we use 45%.
        - Eeff_ICE: the efficiency of the Internal Combustion Engine (ICE) energy conversion system on board, includes ICE and gearbox. This value is approximately 35%.
        - Eeff_Battery: the efficiency of the battery energy conversion system on board. Batteries use 80% capacity to prolong life cycle, and lose efficiency in AC/DC converter, electric motor. Generally this value is between 70% - 95%, here we use 80 %.

        data source:
        Marin report 2019, Energietransitie emissieloze binnenvaart, vooronderzoek ontwerpaspecten, systeem configuraties.(Energy transition zero-emission inland shipping, preliminary research on design aspects, system configurations)
        add other ref

        """
        self.Eeff_FuelCell = 0.45
        self.Eeff_ICE = 0.38
        self.Eeff_Battery = 0.8

    def SFC_general(self):
        """Specific Fuel Consumption (SFC) is calculated by energy density and energy conversion efficiency.
        The SFC calculation equation, SFC = 1 / (energy density * energy conversion efficiency), can be found in the paper of Kim et al (2020)(A Preliminary Study on an Alternative Ship Propulsion System Fueled by Ammonia: Environmental and Economic Assessments, https://doi.org/10.3390/jmse8030183).

        for diesel SFC, there are 3 kinds of general diesel SFC
        - SFC_diesel_ICE_mass, calculated by net diesel gravimetric density and ICE energy-power system efficiency, without considering engine performence variation due to engine ages
        - SFC_diesel_ICE_vol, calculated by net diesel volumetric density and ICE energy-power system efficiency, without considering engine performence variation due to engine ages
        - SFC_diesel_C_year, a group of SFC considering ICE engine performence variation due to engine ages (C_year), based on TNO (2019)

        Please note: later on a correction factor has to be applied to get the total SFC
        """

        # ToDo: Check - in the function 'energy_density' the 'Edens*' are in kWh/kg, but the units in the comments here,
        #  suggest that the 'SFC*' are in g/kWh (wonder if that then should not be kg/kWh)
        # to estimate the requirement of the amount of ZES_batterypacks for different IET scenarios, we include ZES battery capacity per container here.
        # ZES_batterypack capacity > 2000kWh, its average usable energy = 2000 kWh,  mass = 27 ton, vol = 20ft A60 container (6*2.5*2.5 = 37.5 m3) (source: ZES report)
        self.energy_density()
        self.energy_conversion_efficiency()

        self.ZES_batterypack2000kWh = 2000  # kWh/pack,

        # SFC in mass for Fuel Cell engine
        self.SFC_LH2_FuelCell_mass = 1 / (
            self.Edens_LH2_mass * self.Eeff_FuelCell
        )  # g/kWh
        self.SFC_eLNG_FuelCell_mass = 1 / (
            self.Edens_eLNG_mass * self.Eeff_FuelCell
        )  # g/kWh
        self.SFC_eMethanol_FuelCell_mass = 1 / (
            self.Edens_eMethanol_mass * self.Eeff_FuelCell
        )  # g/kWh
        self.SFC_eNH3_FuelCell_mass = 1 / (
            self.Edens_eNH3_mass * self.Eeff_FuelCell
        )  # g/kWh

        # SFC in mass for ICE engine
        self.SFC_diesel_ICE_mass = 1 / (self.Edens_diesel_mass * self.Eeff_ICE)  # g/kWh
        self.SFC_eLNG_ICE_mass = 1 / (self.Edens_eLNG_mass * self.Eeff_ICE)  # g/kWh
        self.SFC_eMethanol_ICE_mass = 1 / (
            self.Edens_eMethanol_mass * self.Eeff_ICE
        )  # g/kWh
        self.SFC_eNH3_ICE_mass = 1 / (self.Edens_eNH3_mass * self.Eeff_ICE)  # g/kWh

        # SFC in mass and volume for battery electric ships
        self.SFC_Li_NMC_Battery_mass = 1 / (
            self.Edens_Li_NMC_Battery_mass * self.Eeff_Battery
        )  # g/kWh
        self.SFC_Li_NMC_Battery_vol = 1 / (
            self.Edens_Li_NMC_Battery_vol * self.Eeff_Battery
        )  # m3/kWh
        self.SFC_ZES_battery2000kWh = 1 / (
            self.ZES_batterypack2000kWh * self.Eeff_Battery
        )  # kWh

        # SFC in volume for Fuel Cell engine
        self.SFC_LH2_FuelCell_vol = 1 / (
            self.Edens_LH2_vol * self.Eeff_FuelCell
        )  # m3/kWh
        self.SFC_eLNG_FuelCell_vol = 1 / (
            self.Edens_eLNG_vol * self.Eeff_FuelCell
        )  # m3/kWh
        self.SFC_eMethanol_FuelCell_vol = 1 / (
            self.Edens_eMethanol_vol * self.Eeff_FuelCell
        )  # m3/kWh
        self.SFC_eNH3_FuelCell_vol = 1 / (
            self.Edens_eNH3_vol * self.Eeff_FuelCell
        )  # m3/kWh

        # SFC in volume for ICE engine
        self.SFC_diesel_ICE_vol = 1 / (self.Edens_diesel_vol * self.Eeff_ICE)  # m3/kWh
        self.SFC_eLNG_ICE_vol = 1 / (self.Edens_eLNG_vol * self.Eeff_ICE)  # m3/kWh
        self.SFC_eMethanol_ICE_vol = 1 / (
            self.Edens_eMethanol_vol * self.Eeff_ICE
        )  # m3/kWh
        self.SFC_eNH3_ICE_vol = 1 / (self.Edens_eNH3_vol * self.Eeff_ICE)  # m3/kWh

        # Another source of diesel SFC: The general diesel SFC (g/kWh) which are based on the construction year of the engine (TNO)

        if self.C_year < 1974:
            self.SFC_diesel_C_year = 235
        if 1975 <= self.C_year <= 1979:
            self.SFC_diesel_C_year = 230
        if 1980 <= self.C_year <= 1984:
            self.SFC_diesel_C_year = 225
        if 1985 <= self.C_year <= 1989:
            self.SFC_diesel_C_year = 220
        if 1990 <= self.C_year <= 1994:
            self.SFC_diesel_C_year = 220
        if 1995 <= self.C_year <= 2002:
            self.SFC_diesel_C_year = 205
        if 2003 <= self.C_year <= 2007:
            self.SFC_diesel_C_year = 200
        if 2008 <= self.C_year <= 2019:
            self.SFC_diesel_C_year = 200
        if self.C_year > 2019:
            if self.L_w == 1:
                self.SFC_diesel_C_year = 205
            else:
                self.SFC_diesel_C_year = 190

        logger.debug(
            f"The general fuel consumption factor for diesel is {self.SFC_diesel_C_year} g/kWh"
        )

    def correction_factors(self, v, h_0):
        """Partial engine load correction factors (C_partial_load):

        - The correction factors have to be multiplied by the general emission factors (or general SFC), to get the total emission factors (or final SFC)
        - The correction factor takes into account the effect of the partial engine load
        - When the partial engine load is low, the correction factors for ICE engine are higher (ICE engine is less efficient at lower enegine load)
        - the correction factors for emissions and diesel fuel in ICE engine are based on literature TNO (2019)
        - For fuel cell enegines(PEMFC & SOFC), the correction factors are lower when the partial engine load is low (fuel cell enegine is more efficient at lower enegine load)
        - the correction factors for renewable fuels used in fuel cell engine are based on literature Kim et al (2020) (A Preliminary Study on an Alternative Ship Propulsion System Fueled by Ammonia: Environmental and Economic Assessments, https://doi.org/10.3390/jmse8030183)
        """
        # TODO: create correction factors for renewable powered ship, the factor may be 100%
        self.calculate_total_power_required(
            v=v, h_0=h_0
        )  # You need the P_partial values

        # Import the correction factors table
        # TODO: use package data, not an arbitrary location
        self.C_partial_load = load_partial_engine_load_correction_factors()
        self.C_partial_load_battery = 1  # assume the battery energy consumption is not influenced by different engine load

        for i in range(20):
            # If the partial engine load is smaller or equal to 5%, the correction factors corresponding to P_partial = 5% are assigned.
            if self.P_partial <= self.C_partial_load.iloc[0, 0]:
                self.C_partial_load_CO2 = self.C_partial_load.iloc[0, 5]
                self.C_partial_load_PM10 = self.C_partial_load.iloc[0, 6]
                self.C_partial_load_fuel_ICE = (
                    self.C_partial_load_CO2
                )  # CO2 emission is generated from fuel consumption, so these two
                # correction factors are equal
                self.C_partial_load_PEMFC = self.C_partial_load.iloc[0, 7]
                self.C_partial_load_SOFC = self.C_partial_load.iloc[0, 8]

                # The NOX correction factors are dependend on the construction year of the engine and the weight class
                if self.C_year < 2008:
                    self.C_partial_load_NOX = self.C_partial_load.iloc[
                        0, 1
                    ]  # <= CCR-1 class
                if 2008 <= self.C_year <= 2019:
                    self.C_partial_load_NOX = self.C_partial_load.iloc[
                        0, 2
                    ]  # CCR-2 / Stage IIIa
                if self.C_year > 2019:
                    if self.L_w == 1:  #
                        self.C_partial_load_NOX = self.C_partial_load.iloc[
                            0, 3
                        ]  # Stage V: IWP/IWA-v/c-3 class (vessels with P <300 kW: assumed to be weight class L1)
                    else:
                        self.C_partial_load_NOX = self.C_partial_load.iloc[
                            0, 4
                        ]  # Stage V:IWP/IWA-v/c-4 class (vessels with P >300 kw: assumed to be weight class L2-L3)

            # If the partial engine load is greater than 5%:
            # It is determined inbetween which two percentages in the table the partial engine load lies
            # The correction factor is determined by means of linear interpolation

            elif (
                self.C_partial_load.iloc[i, 0]
                < self.P_partial
                <= self.C_partial_load.iloc[i + 1, 0]
            ):
                self.C_partial_load_CO2 = (
                    (self.P_partial - self.C_partial_load.iloc[i, 0])
                    * (
                        self.C_partial_load.iloc[i + 1, 5]
                        - self.C_partial_load.iloc[i, 5]
                    )
                ) / (
                    self.C_partial_load.iloc[i + 1, 0] - self.C_partial_load.iloc[i, 0]
                ) + self.C_partial_load.iloc[
                    i, 5
                ]
                self.C_partial_load_PM10 = (
                    (self.P_partial - self.C_partial_load.iloc[i, 0])
                    * (
                        self.C_partial_load.iloc[i + 1, 6]
                        - self.C_partial_load.iloc[i, 6]
                    )
                ) / (
                    self.C_partial_load.iloc[i + 1, 0] - self.C_partial_load.iloc[i, 0]
                ) + self.C_partial_load.iloc[
                    i, 6
                ]
                self.C_partial_load_fuel_ICE = (
                    self.C_partial_load_CO2
                )  # CO2 emission is generated from fuel consumption, so these two
                # correction factors are equal
                self.C_partial_load_PEMFC = (
                    (self.P_partial - self.C_partial_load.iloc[i, 0])
                    * (
                        self.C_partial_load.iloc[i + 1, 7]
                        - self.C_partial_load.iloc[i, 7]
                    )
                ) / (
                    self.C_partial_load.iloc[i + 1, 0] - self.C_partial_load.iloc[i, 0]
                ) + self.C_partial_load.iloc[
                    i, 7
                ]
                self.C_partial_load_SOFC = (
                    (self.P_partial - self.C_partial_load.iloc[i, 0])
                    * (
                        self.C_partial_load.iloc[i + 1, 8]
                        - self.C_partial_load.iloc[i, 8]
                    )
                ) / (
                    self.C_partial_load.iloc[i + 1, 0] - self.C_partial_load.iloc[i, 0]
                ) + self.C_partial_load.iloc[
                    i, 8
                ]
                if self.C_year < 2008:
                    self.C_partial_load_NOX = (
                        (self.P_partial - self.C_partial_load.iloc[i, 0])
                        * (
                            self.C_partial_load.iloc[i + 1, 1]
                            - self.C_partial_load.iloc[i, 1]
                        )
                    ) / (
                        self.C_partial_load.iloc[i + 1, 0]
                        - self.C_partial_load.iloc[i, 0]
                    ) + self.C_partial_load.iloc[
                        i, 1
                    ]
                if 2008 <= self.C_year <= 2019:
                    self.C_partial_load_NOX = (
                        (self.P_partial - self.C_partial_load.iloc[i, 0])
                        * (
                            self.C_partial_load.iloc[i + 1, 2]
                            - self.C_partial_load.iloc[i, 2]
                        )
                    ) / (
                        self.C_partial_load.iloc[i + 1, 0]
                        - self.C_partial_load.iloc[i, 0]
                    ) + self.C_partial_load.iloc[
                        i, 2
                    ]
                if self.C_year > 2019:
                    if self.L_w == 1:
                        self.C_partial_load_NOX = (
                            (self.P_partial - self.C_partial_load.iloc[i, 0])
                            * (
                                self.C_partial_load.iloc[i + 1, 3]
                                - self.C_partial_load.iloc[i, 3]
                            )
                        ) / (
                            self.C_partial_load.iloc[i + 1, 0]
                            - self.C_partial_load.iloc[i, 0]
                        ) + self.C_partial_load.iloc[
                            i, 3
                        ]
                    else:
                        self.C_partial_load_NOX = (
                            (self.P_partial - self.C_partial_load.iloc[i, 0])
                            * (
                                self.C_partial_load.iloc[i + 1, 4]
                                - self.C_partial_load.iloc[i, 4]
                            )
                        ) / (
                            self.C_partial_load.iloc[i + 1, 0]
                            - self.C_partial_load.iloc[i, 0]
                        ) + self.C_partial_load.iloc[
                            i, 4
                        ]

            # If the partial engine load is => 100%, the correction factors corresponding to P_partial = 100% are assigned.
            elif self.P_partial >= self.C_partial_load.iloc[19, 0]:
                self.C_partial_load_CO2 = self.C_partial_load.iloc[19, 5]
                self.C_partial_load_PM10 = self.C_partial_load.iloc[19, 6]
                self.C_partial_load_fuel_ICE = (
                    self.C_partial_load_CO2
                )  # CO2 emission is generated from fuel consumption, so these two
                # correction factors are equal
                self.C_partial_load_PEMFC = self.C_partial_load.iloc[19, 7]
                self.C_partial_load_SOFC = self.C_partial_load.iloc[19, 8]
                # The NOX correction factors are dependend on the construction year of the engine and the weight class
                if self.C_year < 2008:
                    self.C_partial_load_NOX = self.C_partial_load.iloc[
                        19, 1
                    ]  # <= CCR-1 class
                if 2008 <= self.C_year <= 2019:
                    self.C_partial_load_NOX = self.C_partial_load.iloc[
                        19, 2
                    ]  # CCR-2 / Stage IIIa
                if self.C_year > 2019:
                    if self.L_w == 1:  #
                        self.C_partial_load_NOX = self.C_partial_load.iloc[
                            19, 3
                        ]  # Stage V: IWP/IWA-v/c-3 class (vessels with P <300 kW: assumed to be weight class L1)
                    else:
                        self.C_partial_load_NOX = self.C_partial_load.iloc[
                            19, 4
                        ]  # Stage V:IWP/IWA-v/c-4 class (vessels with P >300 kw: assumed to be weight class L2-L3)

        logger.debug(
            f"Partial engine load correction factor of CO2 is {self.C_partial_load_CO2}"
        )
        logger.debug(
            f"Partial engine load correction factor of PM10 is {self.C_partial_load_PM10}"
        )
        logger.debug(
            f"Partial engine load correction factor of NOX is {self.C_partial_load_NOX}"
        )
        logger.debug(
            f"Partial engine load correction factor of diesel fuel consumption in ICE is {self.C_partial_load_fuel_ICE}"
        )
        logger.debug(
            f"Partial engine load correction factor of fuel consumption in PEMFC is {self.C_partial_load_PEMFC}"
        )
        logger.debug(
            f"Partial engine load correction factor of fuel consumption in SOFC is {self.C_partial_load_SOFC}"
        )
        logger.debug(
            f"Partial engine load correction factor of energy consumption in battery is {self.C_partial_load_battery}"
        )

    def calculate_emission_factors_total(self, v, h_0):
        """Total emission factors:

        - The total emission factors can be computed by multiplying the general emission factor by the correction factor
        """

        self.emission_factors_general()  # You need the values of the general emission factors of CO2, PM10, NOX
        self.correction_factors(
            v=v, h_0=h_0
        )  # You need the correction factors of CO2, PM10, NOX

        # The total emission factor is calculated by multiplying the general emission factor (EF_CO2 / EF_PM10 / EF_NOX)
        # by the correction factor (C_partial_load_CO2 / C_partial_load_PM10 / C_partial_load_NOX)

        self.total_factor_CO2 = self.EF_CO2 * self.C_partial_load_CO2
        self.total_factor_PM10 = self.EF_PM10 * self.C_partial_load_PM10
        self.total_factor_NOX = self.EF_NOX * self.C_partial_load_NOX

        logger.debug(
            f"The total emission factor of CO2 is {self.total_factor_CO2} g/kWh"
        )
        logger.debug(
            f"The total emission factor of PM10 is {self.total_factor_PM10} g/kWh"
        )
        logger.debug(f"The total emission factor CO2 is {self.total_factor_NOX} g/kWh")

    def calculate_SFC_final(self, v, h_0):
        """The final SFC is computed by multiplying the general SFC by the partial engine load correction factor.

        The calculation of final SFC below includes
        - the final SFC of LH2, eLNG, eMethanol, eNH3 in mass and volume while using Fuel Cell Engine (PEMFC, SOFC)
        - the final SFC of eLNG, eMethanol, eNH3 in mass and volume while using Internal Combustion Engine
        - the final SFC of diesel in mass and volume while using Internal Combustion Engine
        - the final SFC of battery in mass and volume while use battery-electric power system
        """

        self.SFC_general()  # You need the values of the general SFC
        self.correction_factors(v=v, h_0=h_0)  # You need the correction factors of SFC

        # final SFC of fuel cell in mass   [g/kWh]
        self.final_SFC_LH2_mass_PEMFC = (
            self.SFC_LH2_FuelCell_mass * self.C_partial_load_PEMFC
        )
        self.final_SFC_LH2_mass_SOFC = (
            self.SFC_LH2_FuelCell_mass * self.C_partial_load_SOFC
        )
        self.final_SFC_eLNG_mass_PEMFC = (
            self.SFC_eLNG_FuelCell_mass * self.C_partial_load_PEMFC
        )
        self.final_SFC_eLNG_mass_SOFC = (
            self.SFC_eLNG_FuelCell_mass * self.C_partial_load_SOFC
        )
        self.final_SFC_eMethanol_mass_PEMFC = (
            self.SFC_eMethanol_FuelCell_mass * self.C_partial_load_PEMFC
        )
        self.final_SFC_eMethanol_mass_SOFC = (
            self.SFC_eMethanol_FuelCell_mass * self.C_partial_load_SOFC
        )
        self.final_SFC_eNH3_mass_PEMFC = (
            self.SFC_eNH3_FuelCell_mass * self.C_partial_load_PEMFC
        )
        self.final_SFC_eNH3_mass_SOFC = (
            self.SFC_eNH3_FuelCell_mass * self.C_partial_load_SOFC
        )

        # final SFC of fuel cell in vol  [m3/kWh]
        self.final_SFC_LH2_vol_PEMFC = (
            self.SFC_LH2_FuelCell_vol * self.C_partial_load_PEMFC
        )
        self.final_SFC_LH2_vol_SOFC = (
            self.SFC_LH2_FuelCell_vol * self.C_partial_load_SOFC
        )
        self.final_SFC_eLNG_vol_PEMFC = (
            self.SFC_eLNG_FuelCell_vol * self.C_partial_load_PEMFC
        )
        self.final_SFC_eLNG_vol_SOFC = (
            self.SFC_eLNG_FuelCell_vol * self.C_partial_load_SOFC
        )
        self.final_SFC_eMethanol_vol_PEMFC = (
            self.SFC_eMethanol_FuelCell_vol * self.C_partial_load_PEMFC
        )
        self.final_SFC_eMethanol_vol_SOFC = (
            self.SFC_eMethanol_FuelCell_vol * self.C_partial_load_SOFC
        )
        self.final_SFC_eNH3_vol_PEMFC = (
            self.SFC_eNH3_FuelCell_vol * self.C_partial_load_PEMFC
        )
        self.final_SFC_eNH3_vol_SOFC = (
            self.SFC_eNH3_FuelCell_vol * self.C_partial_load_SOFC
        )

        # final SFC of ICE in mass [g/kWh]
        self.final_SFC_diesel_C_year_ICE_mass = (
            self.SFC_diesel_C_year * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_diesel_ICE_mass = (
            self.SFC_diesel_ICE_mass * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eLNG_ICE_mass = (
            self.SFC_eLNG_ICE_mass * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eMethanol_ICE_mass = (
            self.SFC_eMethanol_ICE_mass * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eNH3_ICE_mass = (
            self.SFC_eNH3_ICE_mass * self.C_partial_load_fuel_ICE
        )

        # final SFC of ICE in vol  [m3/kWh]
        self.final_SFC_diesel_ICE_vol = (
            self.SFC_diesel_ICE_vol * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eLNG_ICE_vol = (
            self.SFC_eLNG_ICE_vol * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eMethanol_ICE_vol = (
            self.SFC_eMethanol_ICE_vol * self.C_partial_load_fuel_ICE
        )
        self.final_SFC_eNH3_ICE_vol = (
            self.SFC_eNH3_ICE_vol * self.C_partial_load_fuel_ICE
        )

        # final SFC of battery in mass and vol
        self.final_SFC_Li_NMC_Battery_mass = (
            self.SFC_Li_NMC_Battery_mass * self.C_partial_load_battery
        )  # g/kWh
        self.final_SFC_Li_NMC_Battery_vol = (
            self.SFC_Li_NMC_Battery_vol * self.C_partial_load_battery
        )  # m3/kWh
        self.final_SFC_Battery2000kWh = (
            self.SFC_ZES_battery2000kWh * self.C_partial_load_battery
        )  # kWh

    def calculate_diesel_use_g_m(self, v):
        """Total diesel fuel use in g/m:

        - The total fuel use in g/m can be computed by total fuel use in g (P_tot * delta_t * self.total_factor_) divided by the sailing distance (v * delt_t)
        """
        self.diesel_use_g_m = (
            self.P_given * self.final_SFC_diesel_ICE_mass / v
        ) / 3600  # without considering C_year
        self.diesel_use_g_m_C_year = (
            self.P_given * self.final_SFC_diesel_C_year_ICE_mass / v
        ) / 3600  # considering C_year

    def calculate_diesel_use_g_s(self):
        """Total diesel fuel use in g/s:

        - The total fuel use in g/s can be computed by total emission in g (P_tot * delta_t * self.total_factor_) diveded by the sailing duration (delt_t)
        """
        self.diesel_use_g_s = (
            self.P_given * self.final_SFC_diesel_ICE_mass / 3600
        )  # without considering C_year
        self.diesel_use_g_s_C_year = (
            self.P_given * self.final_SFC_diesel_C_year_ICE_mass / 3600
        )  # considering C_year

    def calculate_emission_rates_g_m(self, v):
        """CO2, PM10, NOX emission rates in g/m:

        - The CO2, PM10, NOX emission rates in g/m can be computed by total fuel use in g (P_tot * delta_t * self.total_factor_) diveded by the sailing distance (v * delt_t)
        """
        self.emission_g_m_CO2 = self.P_given * self.total_factor_CO2 / v / 3600
        self.emission_g_m_PM10 = self.P_given * self.total_factor_PM10 / v / 3600
        self.emission_g_m_NOX = self.P_given * self.total_factor_NOX / v / 3600

    def calculate_emission_rates_g_s(self):
        """CO2, PM10, NOX emission rates in g/s:

        - The CO2, PM10, NOX emission rates in g/s can be computed by total fuel use in g (P_tot * delta_t * self.total_factor_) diveded by the sailing duration (delt_t)
        """
        self.emission_g_s_CO2 = self.P_given * self.total_factor_CO2 / 3600
        self.emission_g_s_PM10 = self.P_given * self.total_factor_PM10 / 3600
        self.emission_g_s_NOX = self.P_given * self.total_factor_NOX / 3600

    def calculate_max_sinkage(self, v, h_0, width=150):
        """Calculate the maximum sinkage of a moving ship

        the calculation equation is described in Barrass, B. & Derrett, R.'s book (2006), Ship Stability for Masters and Mates,
        chapter 42. https://doi.org/10.1016/B978-0-08-097093-6.00042-6

        some explanation for the variables in the equation:
        - h_0: water depth
        - v: ship velocity relative to the water
        - width: river width, default to 150
        """

        max_sinkage = 0
        if self.h_squat:
            max_sinkage = calculate_max_sinkage(
                v=v,
                h_0=h_0,
                T=self._T,  # TODO: why _T and not T? moreover: T stems from VesselProperties
                B=self.B,
                C_B=self.C_B,
                width=width,
            )

        return max_sinkage

    def calculate_h_squat(self, v, h_0, width=150):
        """Calculate the water depth in case h_squat is set to True

        The amount of water under the keel is calculated h_0 - T. When h_squat is set
        to True, we estimate a max_sinkage that is subtracted from h_0. This values is
        returned as h_squat for further calculation.

        """
        h_squat = h_0 - self.calculate_max_sinkage(v, h_0, width=width)

        return h_squat


class EnergyCalculation:
    """Add information on energy use and effects on energy use."""

    # ToDo: add other alternatives from Marin's table to have completed renewable energy sources
    # ToDo: add renewable fuel cost from Marin's table, add fuel cell / other engine cost, power plan cost to calculate the cost of ship refit or new ships.

    def __init__(self, graph, vessel, *args, **kwargs):
        super().__init__(*args, **kwargs)

        """Initialization"""
        self.graph = graph
        self.vessel = vessel

        self.energy_use = {
            "time_start": [],
            "time_stop": [],
            "edge_start": [],
            "edge_stop": [],
            "P_tot": [],
            "P_given": [],
            "P_installed": [],
            "total_energy": [],
            "total_diesel_consumption_C_year_ICE_mass": [],
            "total_diesel_consumption_ICE_mass": [],
            "total_diesel_consumption_ICE_vol": [],
            "total_LH2_consumption_PEMFC_mass": [],
            "total_LH2_consumption_SOFC_mass": [],
            "total_LH2_consumption_PEMFC_vol": [],
            "total_LH2_consumption_SOFC_vol": [],
            "total_eLNG_consumption_PEMFC_mass": [],
            "total_eLNG_consumption_SOFC_mass": [],
            "total_eLNG_consumption_PEMFC_vol": [],
            "total_eLNG_consumption_SOFC_vol": [],
            "total_eLNG_consumption_ICE_mass": [],
            "total_eLNG_consumption_ICE_vol": [],
            "total_eMethanol_consumption_PEMFC_mass": [],
            "total_eMethanol_consumption_SOFC_mass": [],
            "total_eMethanol_consumption_PEMFC_vol": [],
            "total_eMethanol_consumption_SOFC_vol": [],
            "total_eMethanol_consumption_ICE_mass": [],
            "total_eMethanol_consumption_ICE_vol": [],
            "total_eNH3_consumption_PEMFC_mass": [],
            "total_eNH3_consumption_SOFC_mass": [],
            "total_eNH3_consumption_PEMFC_vol": [],
            "total_eNH3_consumption_SOFC_vol": [],
            "total_eNH3_consumption_ICE_mass": [],
            "total_eNH3_consumption_ICE_vol": [],
            "total_Li_NMC_Battery_mass": [],
            "total_Li_NMC_Battery_vol": [],
            "total_Battery2000kWh_consumption_num": [],
            "total_emission_CO2": [],
            "total_emission_PM10": [],
            "total_emission_NOX": [],
            "stationary": [],
            "water depth": [],
            "distance": [],
            "delta_t": [],
        }

        self.co2_footprint = {"total_footprint": 0, "stationary": 0}
        self.mki_footprint = {"total_footprint": 0, "stationary": 0}

    def calculate_energy_consumption(self):
        """Calculation of energy consumption based on total time in system and properties"""

        # log messages that are related to locking
        # todo: check if this still works with Floors new locking module
        stationary_phase_indicator = [
            "Waiting to enter waiting area stop",  # checked: not sure if still used in locking module
            "Waiting in waiting area stop",  # checked: not sure if still used in locking module
            "Waiting in line-up area stop",  # checked: still used in locking module
            "Passing lock stop",  # checked: still used in locking module
        ]

        # extract relevant elements from the vessel log
        times = [row["Timestamp"] for row in self.vessel.logbook]
        messages = [row["Message"] for row in self.vessel.logbook]
        geometries = [row["Geometry"] for row in self.vessel.logbook]

        # now walk past each logged event (each 'time interval' in the log corresponds to an event)
        for i in range(len(times) - 1):
            # determine the time associated with the logged event (how long did it last)
            delta_t = (times[i + 1] - times[i]).total_seconds()

            if delta_t != 0:
                # append time information to the variables for the dataframe
                self.energy_use["time_start"].append(times[i])
                self.energy_use["time_stop"].append(times[i + 1])

                # append geometry information to the variables for the dataframe
                self.energy_use["edge_start"].append(geometries[i])
                self.energy_use["edge_stop"].append(geometries[i + 1])

                # calculate the distance travelled and the associated velocity
                distance = calculate_distance(geometries[i], geometries[i + 1])
                v = (
                    distance / delta_t
                )  # TODO: this is probably wrong. You don't want the speed over ground here, but the speed to water
                self.energy_use["distance"].append(distance)

                # calculate the delta t
                self.energy_use["delta_t"].append(delta_t)

                # calculate the water depth
                h_0 = calculate_depth(geometries[i], geometries[i + 1], self.graph)

                # printstatements to check the output (can be removed later)
                logger.debug("delta_t: {:.4f} s".format(delta_t))
                logger.debug("distance: {:.4f} m".format(distance))
                logger.debug("velocity: {:.4f} m/s".format(v))

                # we use the calculated velocity to determine the resistance and power required
                # we can switch between the 'original water depth' and 'water depth considering ship squatting' for energy calculation, by using the function "calculate_h_squat (h_squat is set as Yes/No)" in the core.py
                h_0 = self.vessel.calculate_h_squat(v, h_0)
                # print(h_0)
                self.vessel.calculate_total_resistance(v, h_0)
                self.vessel.calculate_total_power_required(v=v, h_0=h_0)

                self.vessel.calculate_emission_factors_total(v=v, h_0=h_0)
                self.vessel.calculate_SFC_final(v=v, h_0=h_0)

                if (
                    messages[i + 1] in stationary_phase_indicator
                ):  # if we are in a stationary stage only log P_hotel
                    # Energy consumed per time step delta_t in the stationary stage
                    energy_delta = self.vessel.P_hotel * delta_t / 3600  # kJ/3600 = kWh

                    # Emissions CO2, PM10 and NOX, in gram - emitted in the stationary stage per time step delta_t,
                    # consuming 'energy_delta' kWh
                    # TODO: check, as it seems that stationary energy use is now not stored.
                    P_hotel_delta = self.vessel.P_hotel  # in kW
                    P_installed_delta = self.vessel.P_installed  # in kW

                else:  # otherwise log P_tot
                    # Energy consumed per time step delta_t in the propulsion stage
                    # TODO: energy_delta should be P_tot times delta_t (was P_given, but then when the vessel is driven with v a strange cutoff occurs, when it is driven by P_tot_given it should be limited by the available power ... that now works)
                    energy_delta = (
                        self.vessel.P_tot * delta_t / 3600
                    )  # kJ/3600 = kWh, when P_tot >= P_installed, P_given = P_installed; when P_tot < P_installed, P_given = P_tot

                    # Emissions CO2, PM10 and NOX, in gram - emitted in the propulsion stage per time step delta_t,
                    # consuming 'energy_delta' kWh
                    P_tot_delta = (
                        self.vessel.P_tot
                    )  # in kW, required power, may exceed installed engine power
                    P_given_delta = self.vessel.P_given  # in kW, actual given power
                    P_installed_delta = self.vessel.P_installed  # in kW
                    emission_delta_CO2 = (
                        self.vessel.total_factor_CO2 * energy_delta
                    )  # Energy consumed per time step delta_t in the                                                                                              #stationary phase # in g
                    emission_delta_PM10 = (
                        self.vessel.total_factor_PM10 * energy_delta
                    )  # in g
                    emission_delta_NOX = (
                        self.vessel.total_factor_NOX * energy_delta
                    )  # in g
                    # Todo: we need to rename the factor name for fuels, not starting with "emission", consider seperating it from emission factors
                    delta_diesel_C_year = (
                        self.vessel.final_SFC_diesel_C_year_ICE_mass * energy_delta
                    )  # in g
                    delta_diesel_ICE_mass = (
                        self.vessel.final_SFC_diesel_ICE_mass * energy_delta
                    )  # in g
                    delta_diesel_ICE_vol = (
                        self.vessel.final_SFC_diesel_ICE_vol * energy_delta
                    )  # in m3

                    delta_LH2_PEMFC_mass = (
                        self.vessel.final_SFC_LH2_mass_PEMFC * energy_delta
                    )  # in g
                    delta_LH2_SOFC_mass = (
                        self.vessel.final_SFC_LH2_mass_SOFC * energy_delta
                    )  # in g
                    delta_LH2_PEMFC_vol = (
                        self.vessel.final_SFC_LH2_vol_PEMFC * energy_delta
                    )  # in m3
                    delta_LH2_SOFC_vol = (
                        self.vessel.final_SFC_LH2_vol_SOFC * energy_delta
                    )  # in m3

                    delta_eLNG_PEMFC_mass = (
                        self.vessel.final_SFC_eLNG_mass_PEMFC * energy_delta
                    )  # in g
                    delta_eLNG_SOFC_mass = (
                        self.vessel.final_SFC_eLNG_mass_SOFC * energy_delta
                    )  # in g
                    delta_eLNG_PEMFC_vol = (
                        self.vessel.final_SFC_eLNG_vol_PEMFC * energy_delta
                    )  # in m3
                    delta_eLNG_SOFC_vol = (
                        self.vessel.final_SFC_eLNG_vol_SOFC * energy_delta
                    )  # in m3
                    delta_eLNG_ICE_mass = (
                        self.vessel.final_SFC_eLNG_ICE_mass * energy_delta
                    )  # in g
                    delta_eLNG_ICE_vol = (
                        self.vessel.final_SFC_eLNG_ICE_vol * energy_delta
                    )  # in m3

                    delta_eMethanol_PEMFC_mass = (
                        self.vessel.final_SFC_eMethanol_mass_PEMFC * energy_delta
                    )  # in g
                    delta_eMethanol_SOFC_mass = (
                        self.vessel.final_SFC_eMethanol_mass_SOFC * energy_delta
                    )  # in g
                    delta_eMethanol_PEMFC_vol = (
                        self.vessel.final_SFC_eMethanol_vol_PEMFC * energy_delta
                    )  # in m3
                    delta_eMethanol_SOFC_vol = (
                        self.vessel.final_SFC_eMethanol_vol_SOFC * energy_delta
                    )  # in m3
                    delta_eMethanol_ICE_mass = (
                        self.vessel.final_SFC_eMethanol_ICE_mass * energy_delta
                    )  # in g
                    delta_eMethanol_ICE_vol = (
                        self.vessel.final_SFC_eMethanol_ICE_vol * energy_delta
                    )  # in m3

                    delta_eNH3_PEMFC_mass = (
                        self.vessel.final_SFC_eNH3_mass_PEMFC * energy_delta
                    )  # in g
                    delta_eNH3_SOFC_mass = (
                        self.vessel.final_SFC_eNH3_mass_SOFC * energy_delta
                    )  # in g
                    delta_eNH3_PEMFC_vol = (
                        self.vessel.final_SFC_eNH3_vol_PEMFC * energy_delta
                    )  # in m3
                    delta_eNH3_SOFC_vol = (
                        self.vessel.final_SFC_eNH3_vol_SOFC * energy_delta
                    )  # in m3
                    delta_eNH3_ICE_mass = (
                        self.vessel.final_SFC_eNH3_ICE_mass * energy_delta
                    )  # in g
                    delta_eNH3_ICE_vol = (
                        self.vessel.final_SFC_eNH3_ICE_vol * energy_delta
                    )  # in m3

                    delta_Li_NMC_Battery_mass = (
                        self.vessel.final_SFC_Li_NMC_Battery_mass * energy_delta
                    )  # in g
                    delta_Li_NMC_Battery_vol = (
                        self.vessel.final_SFC_Li_NMC_Battery_vol * energy_delta
                    )  # in m3
                    delta_Battery2000kWh = (
                        self.vessel.final_SFC_Battery2000kWh * energy_delta
                    )  # in ZESpack number

                    self.energy_use["P_tot"].append(P_tot_delta)
                    self.energy_use["P_given"].append(P_given_delta)
                    self.energy_use["P_installed"].append(P_installed_delta)
                    self.energy_use["total_energy"].append(energy_delta)
                    self.energy_use["stationary"].append(energy_delta)
                    self.energy_use["total_emission_CO2"].append(emission_delta_CO2)
                    self.energy_use["total_emission_PM10"].append(emission_delta_PM10)
                    self.energy_use["total_emission_NOX"].append(emission_delta_NOX)
                    self.energy_use["total_diesel_consumption_C_year_ICE_mass"].append(
                        delta_diesel_C_year
                    )
                    self.energy_use["total_diesel_consumption_ICE_mass"].append(
                        delta_diesel_ICE_mass
                    )
                    self.energy_use["total_diesel_consumption_ICE_vol"].append(
                        delta_diesel_ICE_vol
                    )
                    self.energy_use["total_LH2_consumption_PEMFC_mass"].append(
                        delta_LH2_PEMFC_mass
                    )
                    self.energy_use["total_LH2_consumption_SOFC_mass"].append(
                        delta_LH2_SOFC_mass
                    )
                    self.energy_use["total_LH2_consumption_PEMFC_vol"].append(
                        delta_LH2_PEMFC_vol
                    )
                    self.energy_use["total_LH2_consumption_SOFC_vol"].append(
                        delta_LH2_SOFC_vol
                    )
                    self.energy_use["total_eLNG_consumption_PEMFC_mass"].append(
                        delta_eLNG_PEMFC_mass
                    )
                    self.energy_use["total_eLNG_consumption_SOFC_mass"].append(
                        delta_eLNG_SOFC_mass
                    )
                    self.energy_use["total_eLNG_consumption_PEMFC_vol"].append(
                        delta_eLNG_PEMFC_vol
                    )
                    self.energy_use["total_eLNG_consumption_SOFC_vol"].append(
                        delta_eLNG_SOFC_vol
                    )
                    self.energy_use["total_eLNG_consumption_ICE_mass"].append(
                        delta_eLNG_ICE_mass
                    )
                    self.energy_use["total_eLNG_consumption_ICE_vol"].append(
                        delta_eLNG_ICE_vol
                    )
                    self.energy_use["total_eMethanol_consumption_PEMFC_mass"].append(
                        delta_eMethanol_PEMFC_mass
                    )
                    self.energy_use["total_eMethanol_consumption_SOFC_mass"].append(
                        delta_eMethanol_SOFC_mass
                    )
                    self.energy_use["total_eMethanol_consumption_PEMFC_vol"].append(
                        delta_eMethanol_PEMFC_vol
                    )
                    self.energy_use["total_eMethanol_consumption_SOFC_vol"].append(
                        delta_eMethanol_SOFC_vol
                    )
                    self.energy_use["total_eMethanol_consumption_ICE_mass"].append(
                        delta_eMethanol_ICE_mass
                    )
                    self.energy_use["total_eMethanol_consumption_ICE_vol"].append(
                        delta_eMethanol_ICE_vol
                    )
                    self.energy_use["total_eNH3_consumption_PEMFC_mass"].append(
                        delta_eNH3_PEMFC_mass
                    )
                    self.energy_use["total_eNH3_consumption_SOFC_mass"].append(
                        delta_eNH3_SOFC_mass
                    )
                    self.energy_use["total_eNH3_consumption_PEMFC_vol"].append(
                        delta_eNH3_PEMFC_vol
                    )
                    self.energy_use["total_eNH3_consumption_SOFC_vol"].append(
                        delta_eNH3_SOFC_vol
                    )
                    self.energy_use["total_eNH3_consumption_ICE_mass"].append(
                        delta_eNH3_ICE_mass
                    )
                    self.energy_use["total_eNH3_consumption_ICE_vol"].append(
                        delta_eNH3_ICE_vol
                    )
                    self.energy_use["total_Li_NMC_Battery_mass"].append(
                        delta_Li_NMC_Battery_mass
                    )
                    self.energy_use["total_Li_NMC_Battery_vol"].append(
                        delta_Li_NMC_Battery_vol
                    )
                    self.energy_use["total_Battery2000kWh_consumption_num"].append(
                        delta_Battery2000kWh
                    )

                    self.energy_use["water depth"].append(h_0)
                    # self.energy_use["water depth info from vaarweginformatie.nl"].append(depth)

        # TODO: er moet hier een heel aantal dingen beter worden ingevuld
        # - de kruissnelheid is nu nog per default 1 m/s (zie de Movable mixin). Eigenlijk moet in de
        #   vessel database ook nog een speed_loaded en een speed_unloaded worden toegevoegd.
        # - er zou nog eens goed gekeken moeten worden wat er gedaan kan worden rond kustwerken
        # - en er is nog iets mis met de snelheid rond een sluis

        # - add HasCurrent Class or def
        # - De EnergyCalculation class heeft nu een logboek met veel variabelen. Welke variabelen worden bijgehouden hangt af van een if-statement. De lijsten zijn dus niet allemaal even lang.



# packkage(s) for data handling
from copy import deepcopy
import numpy as np
import pandas as pd


class HasOutput:
    """Support extra output that doesn't fit in the logging module."""

    def __init__(self, *args, **kwargs):
        self.output = {}
        super().__init__(*args, **kwargs)

        if self.__class__.__name__ == 'Vessel' or self.__class__.__name__ == 'IsVessel':

            # Route-dependent output
            self.output["origin"] = ""
            self.output["destination"] = ""
            self.output["route"] = []
            self.output["bound"] = ""
            self.output["anchorage"] = ""
            self.output["turning_basin"] = ""
            self.output["terminal"] = ""
            self.output["berth"] = ""
            self.output["length"] = np.nan
            self.output["beam"] = np.nan
            self.output["draught"] = np.nan
            self.output["sailing_distance"] = np.nan
            self.output["sailing_time"] = np.nan
            self.output['waiting_time'] = {}
            self.output['waiting_time']['Anchorage'] = []
            self.output["waiting_time"]["Terminal"] = []
            self.output["turning_times"] = []
            self.output["(un)loading_times"] = []

            # Edge-dependent output
            self.output["current_node"] = ""
            self.output["next_node"] = ""
            self.output["speed"] = np.nan
            self.output["heading"] = np.nan
            self.output["water_level"] = np.nan
            self.output["MBL"] = np.nan
            self.output["net_ukc"] = np.nan
            self.output["gross_ukc"] = np.nan
            self.output["ship_related_ukc_factors"] = {}

            # Eextra output
            self.output['available_water_depth'] = np.nan
            self.output["required_water_depth"] = np.nan
            self.output["limiting current velocity"] = np.nan

            # Historic output
            self.output["sailed_routes"] = []
            self.output["visited_anchorages"] = []
            self.output["visited_turning_basins"] = []
            self.output["visited_terminals"] = []
            self.output["visited_berths"] = []
            self.output["visited_waiting_areas"] = []
            self.output["visited_lineup_areas"] = []
            self.output["visited_lock_chambers"] = []

        elif self.__class__.__name__ == "IsTerminal" or self.__class__.__name__ == "IsJettyTerminal":
            for berth in [berth.name for berth in self.resource.items]:
                # Visit-dependent output
                self.output[berth] = {}
                self.output[berth]["vessel_information"] = {}
                self.output[berth]["vessel_arrival"] = pd.NaT
                self.output[berth]["vessel_departure"] = pd.NaT
                self.output[berth]["vessel_berthing_times"] = np.nan
                self.output[berth]["vessel_(un)loading_time"] = np.nan
                self.output[berth]["vessel_waiting_time"] = np.nan
                self.output[berth]["visited_vessels"] = []

        elif self.__class__.__name__ == "IsAnchorage":
            self.output["vessel_information"] = {}
            self.output["vessel_arrival"] = pd.NaT
            self.output["vessel_departure"] = pd.NaT
            self.output["vessel_waiting_time"] = np.nan
            self.output["visited_vessels"] = []

        elif self.__class__.__name__ == "IsTurningBasin":
            self.output["vessel_information"] = {}
            self.output["vessel_arrival"] = pd.NaT
            self.output["vessel_departure"] = pd.NaT
            self.output["vessel_turning_time"] = np.nan
            self.output["visited_vessels"] = []

        elif self.__class__.__name__ == "IsLock":
            self.output["visiting_vessels"] = []

        elif self.__class__.__name__ == "IsLockLineUpArea":
            self.output["visiting_vessels"] = []

        elif self.__class__.__name__ == "IsLockWaitingArea":
            self.output["visiting_vessels"] = []

    def update_route_status_report(self, move_stop=False):
        if not move_stop:
            self.output["length"] = self.L
            self.output["beam"] = self.B
            self.output["draught"] = self.T
            self.output["route"] = self.route
            self.output["bound"] = self.bound
            self.output["sailed_routes"].append(self.route)
            self.output["origin"] = self.route[0]
            self.output["destination"] = self.route[-1]
            self.output["sailing_distance"] = 0.0
            self.output["sailing_time"] = 0.0
            for node in self.route:
                if "Anchorage" in self.env.graph.nodes[node]:
                    self.output["anchorage"] = self.env.graph.nodes[node]["Anchorage"][0].name
                if (
                    "Turning basin" in self.env.graph.nodes[node]
                    and self.bound == "inbound"
                    and self.env.graph.nodes[node]["Turning Basin"][0].length <= self.L
                ):
                    self.output["turning_basin"] = self.env.graph.nodes[node]["Turning Basin"][0].name
            if "terminal_of_call" in self.metadata.keys() and self.metadata["terminal_of_call"].size:
                self.output["terminal"] = self.metadata["terminal_of_call"][0]
            if "berth_of_call" in self.metadata.keys() and self.metadata["berth_of_call"].size:
                self.output["berth"] = self.metadata["berth_of_call"][0]

        if move_stop and self.log["Status"]:
            correction = deepcopy(self.log["Status"])[-1]["sailing_distance"]
            if "anchorage" in self.output["sailed_routes"][-1]:
                correction = self.env.vessel_traffic_service.provide_sailing_distance_over_route(
                    self, self.route_after_anchorage[:-1]
                )["Distance"].sum()
            for index, info in enumerate(list(reversed(self.log["Status"]))):
                index = len(self.log["Status"]) - index - 1
                if info["route"] != self.output["sailed_routes"][-1]:
                    break
                if info["bound"] == "inbound":
                    self.log["Status"][index]["sailing_distance"] -= correction
                else:
                    self.log["Status"][index]["sailing_distance"] = -info["sailing_distance"]

    def update_waiting_area_status_report(self, waiting_area, node_waiting_area):
        self.output["visited_waiting_areas"].append(waiting_area.name)
        waiting_area.output["visiting_vessels"] = [user.obj for user in waiting_area.waiting_area[node_waiting_area].users]

    def update_lineup_area_status_report(self, lineup_area, node_lineup_area):
        self.output["visited_lineup_areas"].append(lineup_area.name)
        lineup_area.output["visiting_vessels"] = [user.obj for user in lineup_area.line_up_area[node_lineup_area].users]

    def update_lock_status_report(self, lock):
        self.output["visited_lock_chambers"].append(lock.name)
        lock.output["visiting_vessels"] = [user.obj for user in lock.resource.users]

    def update_turing_basin_status_report(self, turning_basin, turning_stop=False):
        turning_basin.output["vessel_information"] = deepcopy(self.output)
        if not turning_stop:
            turning_basin.output["vessel_arrival"] = self.env.now
            turning_basin.output["vessel_departure"] = self.env.now
            turning_basin.output["vessel_turning_time"] = pd.Timedelta(0, "s")
        if turning_stop:
            self.output["visited_turning_basins"].append(turning_basin.name)
            turning_basin.output["visited_vessels"].append(self.name)
            turning_basin.output["vessel_departure"] = self.env.now
            turning_basin.output["vessel_turning_time"] = (
                deepcopy(turning_basin.output)["vessel_departure"] - deepcopy(turning_basin.output)["vessel_arrival"]
            )
            self.output["turning_times"].append(deepcopy(turning_basin.output)["vessel_turning_time"])

    def update_waiting_status(self, priority=False, availability=False, tidal_window=False, waiting_stop=False):
        if not waiting_stop:
            self.output["waiting_start"] = self.env.now
            if "waiting_time" not in self.output.keys():
                self.output["waiting_time"] = {
                    "Priority": pd.Timedelta(0, "s"),
                    "Availability": pd.Timedelta(0, "s"),
                    "Tidal window": pd.Timedelta(0, "s"),
                }
        else:
            if availability:
                self.output["waiting_time"]["Availability"] = pd.Timedelta(int(self.env.now - self.output["waiting_start"]), "s")
            elif tidal_window:
                self.output["waiting_time"]["Tidal window"] = pd.Timedelta(int(self.env.now - self.output["waiting_start"]), "s")
            elif priority:
                self.output["waiting_time"]["Priority"] = pd.Timedelta(int(self.env.now - self.output["waiting_start"]), "s")

    def update_anchorage_status_report(self, anchorage, departure=False):
        anchorage.output["vessel_information"] = deepcopy(self.output)
        if not departure:
            anchorage.output["vessel_arrival"] = self.env.now
            anchorage.output["vessel_departure"] = self.env.now
            anchorage.output["vessel_waiting_time"] = deepcopy(self.output)["waiting_time"]
        if departure:
            self.output["visited_anchorages"].append(anchorage.name)
            anchorage.output["visited_vessels"].append(self.name)
            anchorage.output["vessel_departure"] = self.env.now
            anchorage.output["vessel_waiting_time"] = deepcopy(self.output)["waiting_time"]
            self.output["waiting_times_in_anchorages"].append(deepcopy(self.output)["waiting_time"])
            try:
                del self.output["waiting_start"]
                del self.output["waiting_time"]
            except KeyError:
                pass

    def update_terminal_berth_status_report(self, terminal, berth):
        terminal.output[berth.name]["vessel_information"] = deepcopy(self.output)

    def update_sailing_status_report(self, current_node, next_node):
        self.output["current_node"] = current_node
        self.output["next_node"] = next_node
        return deepcopy(self.output)
